from contextlib import suppress

import django_filters
from wbcore.filters.mixins import WBCoreFilterMixin
from wbcore.forms import DateRangeField, DateTimeRangeField
from wbcore.utils.date import financial_performance_shortcuts
from wbcore.utils.date_builder.components import Component


class TimeFilter(WBCoreFilterMixin, django_filters.TimeFilter):
    filter_type = "time"


class DateTimeFilter(WBCoreFilterMixin, django_filters.DateTimeFilter):
    filter_type = "datetime"


class DateFilter(WBCoreFilterMixin, django_filters.DateFilter):
    filter_type = "date"


class ShortcutAndPerformanceMixin(WBCoreFilterMixin):
    def __init__(self, shortcuts: list | None = None, performance_mode: bool = False, *args, **kwargs):
        self.shortcuts = shortcuts
        self.performance_mode = performance_mode
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        representation, lookup_expr = super().get_representation(request, name, view)
        representation["performance_mode"] = self.performance_mode

        if self.shortcuts:
            representation["shortcuts"] = self.shortcuts

        return representation, lookup_expr


class DateRangeFilter(ShortcutAndPerformanceMixin, django_filters.Filter):
    field_class = DateRangeField
    filter_type = "daterange"
    default_format = "%Y-%m-%d"

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("lookup_expr", "overlap")
        super().__init__(*args, **kwargs)

    def _get_default(self, *args):
        default = super()._get_default(*args)
        if default is not None:
            lower = upper = None
            if isinstance(default, tuple):
                lower, upper = default

                # if the default is a tuple of components, we need to convert them to string
                if isinstance(lower, Component) and isinstance(upper, Component):
                    return f"{lower},{upper}"

            elif hasattr(default, "lower") and hasattr(default, "upper"):
                lower, upper = default.lower, default.upper
            default = f'{lower.strftime(self.default_format) if lower else ""},{upper.strftime(self.default_format) if upper else ""}'
        return default

    def get_representation(self, request, name, view):
        representation, lookup_expr = super().get_representation(request, name, view)
        representation["lookup_expr"] = {"exact": self.field_name}
        with suppress(KeyError):  # TODO frontend needs to support both exact and overlaps lookup
            default = representation["default"].pop(self.lookup_expr)
            representation["default"]["exact"] = default

        return representation, lookup_expr

    @classmethod
    def base_date_range_filter_method(cls, queryset, field_name, value):
        if value:
            filters = {}
            if value.lower:
                filters[f"{field_name}__gte"] = value.lower
            if value.upper:
                filters[f"{field_name}__lte"] = value.upper
            return queryset.filter(**filters)
        return queryset


class FinancialPerformanceDateRangeFilter(DateRangeFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(performance_mode=True, shortcuts=financial_performance_shortcuts, *args, **kwargs)


class DateTimeRangeFilter(DateRangeFilter):
    field_class = DateTimeRangeField
    default_format = "%Y-%m-%dT%H:%M:%S%z"



from django.utils.translation import gettext as _
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class TeamButtonConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self) -> set:
        return {
            bt.DropDownButton(
                label=_("Website & Coach"),
                icon=WBIcon.UNFOLD.icon,
                buttons=(
                    bt.HyperlinkButton(key="website", label=_("Homepage"), icon=WBIcon.LINK.icon, weight=1),
                    bt.WidgetButton(
                        label=_("Coach"),
                        icon=WBIcon.PERSON.icon,
                        key="coach",
                        weight=2,
                    ),
                ),
            )
        }



from random import randint

import pytest
from wbcore.metadata.configs.display.instance_display.styles import Style
from wbcore.metadata.configs.display.instance_display.utils import (
    grid_definition,
    repeat,
    repeat_field,
    split_list_into_grid_template_area_sublists,
)


def test_repeat():
    assert repeat(2, Style.px(20)) == "repeat(2, 20px)"


def test_repeat_field():
    assert repeat_field(2, Style.px(20)) == "20px 20px"


def test_grid_definition():
    assert grid_definition(Style.px(20), Style.px(10), Style.px(20)) == "20px 10px 20px"


@pytest.mark.parametrize(
    ("input", "output"),
    [
        ([], [[".", ".", "."]]),
        (["abc", "d"], [["abc", "d", "."]]),
        (["abc", "d", "ef"], [["abc", "d", "ef"]]),
        (["dh", "faf", "fsefs", "ss"], [["dh", "faf", "fsefs"], ["ss", ".", "."]]),
        (["dh", "faf", "fsefs", "ss", "rty", "dthf"], [["dh", "faf", "fsefs"], ["ss", "rty", "dthf"]]),
    ],
)
def test_split_list_into_grid_template_area_sublists(input, output):
    assert split_list_into_grid_template_area_sublists(input, 3) == output


def test_split_list_into_grid_template_area_sublists_error():
    with pytest.raises(ValueError):
        split_list_into_grid_template_area_sublists(["abc"], -3)


def test_split_list_into_grid_template_area_sublists_random():
    column_number = randint(1, 5)
    field_list = ["Test" for _ in range(0, 15)]
    for sublist in split_list_into_grid_template_area_sublists(field_list, column_number):
        assert len(sublist) == column_number



from rest_framework.request import Request
from rest_framework.reverse import reverse
from wbcore.configs.decorators import register_config


@register_config
def profile_config(request: Request) -> tuple[str, str]:
    return "profile", reverse("wbcore:profile", request=request)



from functools import cached_property
from typing import Any

import pandas as pd
from django.core.cache import cache
from wbcore.utils.strings import camel_to_snake_case


class CacheMixin:
    extra_cache_kwargs: dict[str, Any]
    CACHE_EMPTY_VALUE = pd.DataFrame()
    CACHE_MAIN_KEY: str = "df"
    CACHE_METHOD: str = "_get_dataframe"

    @cached_property
    def cache_enabled(self) -> bool:
        return getattr(self, "CACHE_ENABLED", False)

    def _get_cache_timeout(self) -> int | None:
        cache_timeout = getattr(self, "CACHE_TIMEOUT", None)
        if callable(cache_timeout):
            return cache_timeout()
        return cache_timeout

    def _get_cache_key(self) -> str:
        cache_key = camel_to_snake_case(self.__class__.__name__)
        if cache_key_prefix := getattr(self, "CACHE_KEY_PREFIX", None):
            if callable(cache_key_prefix):
                cache_key_prefix = cache_key_prefix()
            cache_key += "-" + cache_key_prefix
        return cache_key

    def deserialize_cache_results(self, cached_res: dict[str, Any]) -> Any:
        res = cached_res.pop(self.CACHE_MAIN_KEY, self.CACHE_EMPTY_VALUE)
        # we loop over the other cached kwargs and set them as class attributes
        for k, v in cached_res.items():
            setattr(self, k, v)
        return res

    def serialize_cache_results(self, res: Any) -> dict[str, Any]:
        cached_res = {self.CACHE_MAIN_KEY: res}
        cached_res.update(self.extra_cache_kwargs)
        return cached_res

    def _get_cached_res(self) -> Any:
        if cached_res := cache.get(self._get_cache_key()):
            return self.deserialize_cache_results(cached_res)
        return self.CACHE_EMPTY_VALUE



from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from rest_framework.serializers import ValidationError
from slugify import slugify
from wbcore import serializers
from wbcore import serializers as wb_serializers

from ..models import (
    ClientManagerRelationship,
    Company,
    EmployerEmployeeRelationship,
    Entry,
    Person,
    Position,
    Relationship,
    RelationshipType,
)
from .entries import (
    CompanyRepresentationSerializer,
    EntryRepresentationSerializer,
    PersonRepresentationSerializer,
)


class ClientManagerRelationshipRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:directory:clientmanagerrelationship-detail")

    class Meta:
        model = ClientManagerRelationship
        fields = (
            "id",
            "computed_str",
            "_detail",
        )


class RelationshipTypeRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = RelationshipType
        fields = ("id", "counter_relationship", "title")


class RelationshipRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = Relationship
        fields = (
            "id",
            "computed_str",
        )


class PositionRepresentationSerializer(wb_serializers.RepresentationSerializer):
    endpoint = "wbcore:directory:positionrepresentation-list"
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:directory:position-detail")

    class Meta:
        model = Position
        fields = (
            "id",
            "title",
            "_detail",
        )


class PositionModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = Position
        fields = (
            "id",
            "title",
        )

    def validate(self, data):
        title = data.get("title", None)
        if title:
            position = Position.objects.filter(slugify_title=slugify(title, separator=" "))
            if obj := self.instance:
                position = position.exclude(id=obj.id)
            if position.exists():
                raise ValidationError({"title": _("Cannot add a duplicate position.")})
        return data


class RelationshipTypeModelSerializer(serializers.ModelSerializer):
    _counter_relationship = RelationshipTypeRepresentationSerializer(source="counter_relationship")

    class Meta:
        model = RelationshipType
        fields = (
            "id",
            "title",
            "counter_relationship",
            "_counter_relationship",
        )

    def validate(self, data):
        title = data.get("title", None)
        if title:
            type = RelationshipType.objects.filter(slugify_title=slugify(title, separator=" "))
            if obj := self.instance:
                type = type.exclude(id=obj.id)
            if type.exists():
                raise ValidationError({"title": _("Cannot add duplicate relationship type.")})
        return data


class RelationshipModelSerializer(serializers.ModelSerializer):
    _relationship_type = RelationshipTypeRepresentationSerializer(source="relationship_type")
    from_entry = serializers.PrimaryKeyRelatedField(required=False, queryset=Entry.objects.all())
    _from_entry = EntryRepresentationSerializer(source="from_entry")
    _to_entry = EntryRepresentationSerializer(source="to_entry")

    class Meta:
        model = Relationship
        fields = (
            "id",
            "relationship_type",
            "_relationship_type",
            "from_entry",
            "_from_entry",
            "to_entry",
            "_to_entry",
        )

    def validate(self, data):
        try:
            if data.get("from_entry"):
                data["from_entry_id"] = data["from_entry"].id
            elif self.instance and self.instance.from_entry:
                data["from_entry_id"] = self.instance.from_entry.id
            else:
                data["from_entry_id"] = self.context["view"].kwargs["entry_id"]
        except KeyError:
            raise ValidationError(_("From entry has to be set."))
        from_entry = data.get("from_entry", getattr(self.instance, "from_entry", None))
        to_entry = data.get("to_entry", getattr(self.instance, "to_entry", None))

        if from_entry == to_entry:
            msg = _("Cannot add a duplicate relationship type.")
            raise ValidationError({"from_entry": msg, "to_entry": msg})
        return super().validate(data)


class EmployerEmployeeRelationshipSerializer(serializers.ModelSerializer):
    _employer = CompanyRepresentationSerializer(source="employer")
    _employee = PersonRepresentationSerializer(source="employee")
    _position = PositionRepresentationSerializer(source="position")

    employer_profile_pic = serializers.ImageField(
        source="employer.profile_image", label=_("Profile Picture"), read_only=True
    )
    employee_profile_pic = serializers.ImageField(
        source="employee.profile_image", label=_("Profile Picture"), read_only=True
    )

    class Meta:
        model = EmployerEmployeeRelationship
        fields = (
            "id",
            "primary",
            "employer",
            "_employer",
            "employee",
            "_employee",
            "primary",
            "position",
            "_position",
            "employer_profile_pic",
            "employee_profile_pic",
            "position_name",
        )
        dependency_map = {
            "position_name": ["position"],
        }

    def validate(self, data: dict) -> dict:
        employee: Person | None = data.get("employee", None)
        employer: Company | None = data.get("employer", None)
        primary: bool | None = data.get("primary", None)

        if employee and employer and not self.instance:
            if employer in employee.employers.all():
                raise ValidationError(
                    {
                        "employee": _("{employee} is already an employee of {employer}.").format(
                            employee=employee.computed_str, employer=employer.computed_str
                        )
                    }
                )
        elif primary is not None and self.instance:
            if (
                EmployerEmployeeRelationship.objects.exclude(id=self.instance.id)
                .filter(employer=self.instance.employer, employee=self.instance.employee, primary=True)
                .exists()
                and primary is False
            ):
                raise ValidationError(
                    {"primary": _("Cannot degrade primary employer. Add a new primary employer instead.")}
                )
        return data


class ClientManagerModelSerializer(serializers.ModelSerializer):
    client = serializers.PrimaryKeyRelatedField(
        queryset=Entry.objects.all(),
        default=serializers.DefaultFromGET("client"),
        label=gettext_lazy("Client"),
        read_only=lambda view: not view.is_modifiable,
    )
    _client = EntryRepresentationSerializer(many=False, source="client")
    relationship_manager = serializers.PrimaryKeyRelatedField(
        queryset=Person.objects.all(),
        default=serializers.DefaultFromGET("relationship_manager"),
        label=gettext_lazy("Relationship Manager"),
        read_only=lambda view: not view.is_modifiable,
    )
    primary = wb_serializers.BooleanField(read_only=lambda view: not view.is_modifiable)
    _relationship_manager = PersonRepresentationSerializer(many=False, source="relationship_manager")

    def validate(self, data):
        client = data.get("client", self.instance.client if self.instance else None)
        relationship_manager = data.get(
            "relationship_manager", self.instance.relationship_manager if self.instance else None
        )
        primary = data.get("primary", self.instance.primary if self.instance else None)

        if not client and not relationship_manager:
            raise ValidationError(
                {"client": _("You need to select a client"), "relationship_manager": _("You need to select a manager")}
            )

        if not client:
            raise ValidationError({"client": _("You need to select a client")})

        if not relationship_manager:
            raise ValidationError({"relationship_manager": _("You need to select a manager")})

        if client and relationship_manager:
            is_primary_manager = ClientManagerRelationship.objects.filter(
                client=client, relationship_manager=relationship_manager, primary=True
            ).exists()

            qs = ClientManagerRelationship.objects.filter(
                relationship_manager=relationship_manager,
                client=client,
                status=ClientManagerRelationship.Status.PENDINGADD,
            )
            if self.instance:
                qs = qs.exclude(id=self.instance.id)
            if qs.exists():
                raise ValidationError(
                    {
                        "non_field_errors": _(
                            "There is already a pending request to put {manager} in charge of {client}."
                        ).format(manager=str(relationship_manager), client=client.computed_str)
                    }
                )

            qs = ClientManagerRelationship.objects.filter(
                relationship_manager=relationship_manager, client=client, status=ClientManagerRelationship.Status.DRAFT
            )
            if self.instance:
                qs = qs.exclude(id=self.instance.id)
            if qs.exists():
                raise ValidationError(
                    {
                        "non_field_errors": _("A draft to put {manager} in charge of {client} already exists.").format(
                            manager=str(relationship_manager), client=client.computed_str
                        )
                    }
                )
            if relationship_manager.id == client.id:
                raise ValidationError(
                    {"relationship_manager": _("Client and relationship manager cannot be the same person.")}
                )
            if relationship_manager in client.relationship_managers.all() and not primary and is_primary_manager:
                raise ValidationError(
                    {
                        "primary": _(
                            "Cannot degrade primary manager. Make a primary request for a different manager instead."
                        )
                    }
                )

            qs = ClientManagerRelationship.objects.filter(
                relationship_manager=relationship_manager,
                client=client,
                status__in=[ClientManagerRelationship.Status.APPROVED, ClientManagerRelationship.Status.PENDINGREMOVE],
            )
            if self.instance:
                qs = qs.exclude(id=self.instance.id)
            if qs.exists():
                raise ValidationError(
                    {
                        "non_field_errors": _("{manager} is already in charge of {client}.").format(
                            manager=str(relationship_manager), client=client.computed_str
                        )
                    }
                )
        return data

    class Meta:
        model = ClientManagerRelationship
        fields = (
            "id",
            "status",
            "client",
            "_client",
            "relationship_manager",
            "_relationship_manager",
            "created",
            "primary",
            "_additional_resources",
        )
        read_only_fields = ("created",)


class UserIsClientModelSerializer(wb_serializers.ModelSerializer):
    relationship_manager_name = wb_serializers.CharField(read_only=True, label="Name")
    relationship_manager_email = wb_serializers.CharField(read_only=True, label="Email")
    relationship_manager_phone_number = wb_serializers.CharField(read_only=True, label="Phone")

    class Meta:
        model = ClientManagerRelationship
        fields = (
            "id",
            "primary",
            "relationship_manager",
            "relationship_manager_name",
            "relationship_manager_email",
            "relationship_manager_phone_number",
        )



import pytest
from wbcore.contrib.example_app.factories import (
    EventFactory,
    EventTypeFactory,
    LeagueFactory,
)
from wbcore.contrib.example_app.utils import get_event_types_for_league


@pytest.mark.django_db
class TestUtils:
    def test_get_event_types_for_league(self):
        league = LeagueFactory()
        event_type1 = EventTypeFactory(sport=league.sport)
        event_type2 = EventTypeFactory(sport=league.sport)
        event_type3 = EventTypeFactory()
        event_type4 = EventTypeFactory(sport=league.sport)
        EventFactory(event_type=event_type1)
        EventFactory(event_type=event_type2)
        EventFactory(event_type=event_type3)
        league_event_types = get_event_types_for_league(league.pk)
        for event_type in league_event_types:
            assert event_type["name"] in [event_type1.name, event_type2.name]
            assert event_type["name"] not in [event_type3.name, event_type4.name]
            assert " " not in event_type["slugified_name"]



LOOKUP_EXPRESSION_MAPPING = {
    "exact": ("Equals", "="),
    "iexact": ("Equals (case-insensitive)", "≈"),
    "contains": ("Contains", "[=]"),
    "icontains": ("Contains (case-insensitive)", "[≈]"),
    "startswith": ("Starts with", "=]"),
    "istartswith": ("Starts with (case-insensitive)", "≈]"),
    "endswith": ("Ends with", "[="),
    "iendswith": ("Ends with (case-insensitive)", "[≈"),
    "regex": ("Regex", "r="),
    "iregex": ("Regex (case-insensitive)", "r≈"),
    "in": ("In", "∈"),
    "gte": ("Greater than or Equal", ">="),
    "lte": ("Less than or Equal", "<="),
    "gt": ("Greater than", ">"),
    "lt": ("Less than", "<"),
    "overlap": ("Overlaps", "="),
    "isnull": ("Is Null", "Is Null"),
}


ALL_TEXT_LOOKUPS = [
    "exact",
    "iexact",
    "contains",
    "icontains",
    "startswith",
    "istartswith",
    "endswith",
    "iendswith",
    "regex",
    "iregex",
]


def get_lookup_label(lookup_expr: str) -> str:
    return LOOKUP_EXPRESSION_MAPPING.get(lookup_expr, (lookup_expr, lookup_expr))[0]


def get_lookup_icon(lookup_expr: str) -> str:
    return LOOKUP_EXPRESSION_MAPPING.get(lookup_expr, (lookup_expr, lookup_expr))[1]



from dataclasses import dataclass

from django.utils.translation import gettext as _
from rest_framework.response import Response
from wbcore.metadata.configs.buttons import ButtonConfig
from wbcore.metadata.configs.display.instance_display import Display


@dataclass
class PreAction:
    message: str
    instance_display: Display | None = None
    confirm_config: ButtonConfig = ButtonConfig(label=_("Confirm"), title=_("Confirm"))
    cancel_config: ButtonConfig = ButtonConfig(label=_("Cancel"), title=_("Cancel"))

    def to_response(self):
        response_dict = {
            "message": self.message,
            "confirm_config": dict(self.confirm_config),
            "cancel_config": dict(self.cancel_config),
        }

        if self.instance_display:
            response_dict["display"] = self.instance_display.serialize()

        return Response(response_dict)



import re
from operator import add, sub

import pytest
from wbcore.utils.date_builder import *


class TestDateBuilder:
    @pytest.mark.parametrize(
        "component,name",
        [
            (SecondStart, "second_start"),
            (SecondEnd, "second_end"),
            (MinuteStart, "minute_start"),
            (MinuteEnd, "minute_end"),
            (HourStart, "hour_start"),
            (HourEnd, "hour_end"),
            (WeekStart, "week_start"),
            (WeekEnd, "week_end"),
            (MonthStart, "month_start"),
            (MonthEnd, "month_end"),
            (QuarterStart, "quarter_start"),
            (QuarterEnd, "quarter_end"),
            (YearStart, "year_start"),
            (YearEnd, "year_end"),
        ],
    )
    def test_component_str(self, component, name):
        assert str(component) == name

    @pytest.mark.parametrize(
        "offset, name",
        [
            (Second, "seconds"),
            (Minute, "minutes"),
            (Hour, "hours"),
            (Day, "days"),
            (BusinessDay, "bdays"),
            (Week, "weeks"),
            (Month, "months"),
            (Quarter, "quarters"),
            (Year, "years"),
        ],
    )
    @pytest.mark.parametrize("amount", [None, 2])
    def test_offset_str(self, offset, name, amount):
        if amount:
            assert str(offset(amount)) == f"{amount}{name}"
        else:
            assert str(offset) == f"1{name}"

    @pytest.mark.parametrize(
        "component",
        [
            SecondStart,
            SecondEnd,
            MinuteStart,
            MinuteEnd,
            HourStart,
            HourEnd,
            WeekStart,
            WeekEnd,
            MonthStart,
            MonthEnd,
            QuarterStart,
            QuarterEnd,
            YearStart,
            YearEnd,
        ],
    )
    @pytest.mark.parametrize("offset", [Second, Minute, Hour, Day, BusinessDay, Week, Month, Quarter, Year])
    @pytest.mark.parametrize("op", [add, sub])
    def test_component_with_offset(self, component, offset, op):
        results = re.findall("([a-z_]*)([+-]{1})([0-9a-z]*)", str(op(component, offset)))
        assert (
            len(results[0]) == 3
        )  # We expect three groups to be found, 1 for the component, 1 for the operator, and 1 for the offset

    @pytest.mark.parametrize(
        "component",
        [
            (SecondStart),
            (SecondEnd),
            (MinuteStart),
            (MinuteEnd),
            (HourStart),
            (HourEnd),
            (WeekStart),
            (WeekEnd),
            (MonthStart),
            (MonthEnd),
            (QuarterStart),
            (QuarterEnd),
            (YearStart),
            (YearEnd),
        ],
    )
    def test_repr(self, component):
        assert str(component) == repr(component)



from selenium.webdriver.remote.webdriver import WebDriver
from wbcore.contrib.example_app.factories import (
    LeagueFactory,
    SportPersonFactory,
    TeamFactory,
)
from wbcore.contrib.example_app.models import League, Sport, Team
from wbcore.contrib.example_app.serializers import (
    LeagueModelSerializer,
    SportPersonModelSerializer,
    TeamModelSerializer,
)
from wbcore.test import (
    click_element_by_path,
    fill_out_form_fields,
    open_create_instance,
)


def create_new_league_instance(
    driver: WebDriver, field_list: list[str], sport: Sport, is_create_instance_open=True
) -> League:
    if not is_create_instance_open:
        open_create_instance(driver, "Sports App", create_instance_title="Create League")

    league: dict = LeagueFactory.build(sport=sport)
    serializer = LeagueModelSerializer(league)
    fill_out_form_fields(driver, serializer, field_list, league)
    click_element_by_path(driver, "//button[@label='Save and close']")
    return league


def create_new_team_instance(driver: WebDriver, field_list: list[str], is_create_instance_open=True) -> Team:
    if not is_create_instance_open:
        open_create_instance(driver, "Sports App", create_instance_title="Create Team")

    team: dict = TeamFactory.build()
    serializer = TeamModelSerializer(team)
    fill_out_form_fields(driver, serializer, field_list, team)
    click_element_by_path(driver, "//button[@label='Save and close']")
    return team


def create_new_person_instance(driver: WebDriver, field_list: list[str], is_create_instance_open=True):
    if not is_create_instance_open:
        open_create_instance(driver, "Sports App", "Persons", create_instance_title="Create Person")
    person: dict = SportPersonFactory.build()
    serializer = SportPersonModelSerializer(person)
    fill_out_form_fields(driver, serializer, field_list, person)
    click_element_by_path(driver, "//button[@label='Save and close']")
    return person



from django.db import models


class PermissionMixin(models.Model):
    class Meta:
        abstract = True

    @classmethod
    @property
    def view_perm_str(cls) -> str:
        """
        Get the view string permission identifer

        Returns:
            The view string permission identifier
        """
        return f"{cls._meta.app_label}.view_{cls._meta.model_name}"

    @classmethod
    @property
    def change_perm_str(cls) -> str:
        """
        Get the change string permission identifer

        Returns:
            The change string permission identifier
        """
        return f"{cls._meta.app_label}.change_{cls._meta.model_name}"

    @classmethod
    @property
    def delete_perm_str(cls) -> str:
        """
        Get the delete string permission identifer

        Returns:
            The delete string permission identifier
        """
        return f"{cls._meta.app_label}.delete_{cls._meta.model_name}"

    @classmethod
    @property
    def select_perm_str(cls) -> str:
        """
        Get the select string permission identifer

        Returns:
            The select string permission identifier
        """
        return f"{cls._meta.app_label}.select_{cls._meta.model_name}"

    @classmethod
    @property
    def admin_perm_str(cls) -> str:
        """
        Get the admin string permission identifer

        Returns:
            The admin string permission identifier
        """
        return f"{cls._meta.app_label}.administrate_{cls._meta.model_name}"

    @classmethod
    @property
    def add_perm_str(cls) -> str:
        """
        Get the add string permission identifer

        Returns:
            The add string permission identifier
        """
        return f"{cls._meta.app_label}.add_{cls._meta.model_name}"



from django.apps import apps
from wbcore.utils.itertools import get_inheriting_subclasses


def receiver_all_subclasses(signal, sender, *args, **kwargs):
    def _decorator(func):
        signal.connect(func, *args, sender=sender, dispatch_uid=func.__name__ + "_" + sender.__name__, **kwargs)
        for subclass in sender.__subclasses__():
            signal.connect(
                func, *args, sender=subclass, dispatch_uid=func.__name__ + "_" + subclass.__name__, **kwargs
            )

        return func

    return _decorator


def receiver_subclasses(signal, sender, *args, only_leaf=True, **kwargs):
    """
    A decorator for connecting receivers and all receiver's subclasses to signals. Used by passing in the
    signal and keyword arguments to connect::
    The signal needs to be registered after all child models are registered (e.g. by placing the signal in a dedicated
    signal file)
        @receiver_subclasses(post_save, sender=MyModel)
        def signal_receiver(sender, **kwargs):
            ...
    """

    def _decorator(func):
        if isinstance(sender, str):
            sender_class = apps.get_model(sender)
        else:
            sender_class = sender
        for snd in get_inheriting_subclasses(sender_class, only_leaf=only_leaf):
            signal.connect(func, *args, sender=snd, dispatch_uid=func.__name__ + "_" + snd.__name__, **kwargs)
        return func

    return _decorator


def receiver_inherited_through_models(signal, sender, through_field_name, **kwargs):
    """
    A decorator for connecting receivers and all subclasses's through model'.
    """

    def _decorator(func):
        for sub_class in sender.__subclasses__():
            if (through_field := getattr(sub_class, through_field_name, None)) and (
                through_class := getattr(through_field, "through", None)
            ):
                signal.connect(
                    func, sender=through_class, dispatch_uid=func.__name__ + "_" + through_class.__name__, **kwargs
                )

    return _decorator



from dynamic_preferences.registries import global_preferences_registry


def get_timedelta_import_currency_fx_rates():
    return global_preferences_registry.manager()["currency__timedelta_import_currency_fx_rates"]



from wbcore.metadata.configs.titles import TitleViewConfig


class GeographyTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Geographies"

    def get_create_title(self):
        return "New Geography"

    def get_instance_title(self):
        if "pk" in self.view.kwargs:
            return f"Geography: {self.view.get_object()}"
        return "Geography"



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Layout,
    Page,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default


class ConditionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="transition", label=_("Transition")),
            dp.Field(key="attribute_name", label=_("Attribute Name")),
            dp.Field(key="operator", label=_("Operator")),
            dp.Field(key="negate_operator", label=_("Negate Operator")),
            dp.Field(key="expected_value", label=_("Expected Value")),
        ]

        if "transition_id" in self.view.kwargs:
            fields.pop(0)

        return dp.ListDisplay(
            fields=fields,
        )

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["transition", "attribute_name"],
                                ["operator", "negate_operator"],
                                ["expected_value", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                    },
                ),
            ]
        )



from contextlib import suppress

from django.contrib.auth import get_user_model
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from guardian.utils import get_anonymous_user
from wbcore.models import WBModel


class NotificationType(WBModel):
    code = models.CharField(max_length=128, unique=True)
    title = models.CharField(max_length=128, default="")
    help_text = models.CharField(max_length=512, default="")

    stale = models.BooleanField(default=False)
    contenttype = models.ForeignKey(
        to="contenttypes.ContentType", related_name="+", null=True, blank=True, on_delete=models.SET_NULL
    )

    default_enable_web = models.BooleanField(default=False)
    default_enable_mobile = models.BooleanField(default=False)
    default_enable_email = models.BooleanField(default=False)

    def __str__(self) -> str:
        return f"{self.title}"

    class Meta:
        verbose_name = "Notification Type"
        verbose_name_plural = "Notification Types"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:notifications:notification_type"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:notifications:notification_type_representation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"


class NotificationTypeSetting(WBModel):
    notification_type = models.ForeignKey(
        to="notifications.NotificationType", related_name="user_settings", on_delete=models.CASCADE
    )
    user = models.ForeignKey(
        to=get_user_model(), related_name="wbnotification_user_settings", on_delete=models.CASCADE
    )

    enable_web = models.BooleanField(default=False)
    enable_mobile = models.BooleanField(default=False)
    enable_email = models.BooleanField(default=False)

    def __str__(self) -> str:
        return f"{self.user}: {self.notification_type} ({self.enable_web}/{self.enable_mobile}/{self.enable_email})"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:notifications:notification_type_setting"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:notifications:notification_type_setting_representation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{notification_type}}"


@receiver(post_save, sender="notifications.NotificationType")
def post_save_notification_type(instance, **kwargs):
    anonymous_user = get_anonymous_user()
    for user in get_user_model().objects.filter(~models.Q(pk=anonymous_user.pk)):
        NotificationTypeSetting.objects.get_or_create(
            notification_type=instance,
            user=user,
            defaults={
                "enable_web": instance.default_enable_web,
                "enable_mobile": instance.default_enable_mobile,
                "enable_email": instance.default_enable_email,
            },
        )


@receiver(post_save, sender=get_user_model())
def post_save_user(instance, **kwargs):
    with suppress(get_user_model().DoesNotExist):
        anonymous_user = get_anonymous_user()
        if instance.pk == anonymous_user.pk:
            return

    for notification_type in NotificationType.objects.all():
        NotificationTypeSetting.objects.get_or_create(
            notification_type=notification_type,
            user=instance,
            defaults={
                "enable_web": notification_type.default_enable_web,
                "enable_mobile": notification_type.default_enable_mobile,
                "enable_email": notification_type.default_enable_email,
            },
        )



import pytest
from wbcore.contrib.notifications.serializers import (
    NotificationModelSerializer,
    NotificationTypeRepresentationSerializer,
)


@pytest.mark.django_db
class TestNotificationModelSerializer:
    def test_serialize_instance(self, notification, request_with_user):
        serializer = NotificationModelSerializer(notification, context={"request": request_with_user})
        assert serializer.data == {
            "id": notification.id,
            "title": notification.title,
            "notification_type": notification.notification_type.id,
            "_notification_type": NotificationTypeRepresentationSerializer(notification.notification_type).data,
            "body": notification.body,
            "user": notification.user.pk,
            "endpoint": notification.endpoint,
            "sent": notification.sent,
            "read": notification.read,
            "_additional_resources": {},
        }



from contextlib import suppress

from django.core.exceptions import FieldDoesNotExist, FieldError
from django.db import models
from django.db.models.signals import post_delete, pre_delete
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from ordered_model.models import OrderedModelManager
from rest_framework.fields import get_attribute
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.color.fields import ColorField
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.icons.models import IconField
from wbcore.utils.enum import ChoiceEnum


def get_and_update_or_create(model, filter_params, defaults):
    """get or create with default values applied to existing instances

    {model}.objects.get_or_create(*{filter_params}, defaults={defaults})

    Arguments:
        model {django.db.models.Model} -- The model the queryset gets applied to
        filter_params {dict} -- The parameters that the queryset gets filter against
        defaults {dict} -- The default values that will be applied to the instance (create, update)

    Returns:
        django.db.models.Model -- The created/updated instance of the model
    """

    instance, created = model.objects.get_or_create(**filter_params, defaults=defaults)
    if not created:
        model.objects.filter(id=instance.id).update(**defaults)
        instance = model.objects.get(id=instance.id)
        instance.save()
    return instance


class NoGroupBySubquery(models.Subquery):
    """Same as the default django subquery, however does not perform a group by"""

    def get_group_by_cols(self):
        return []


class NoGroupByExpressionWrapper(models.ExpressionWrapper):
    def get_group_by_cols(self):
        return []


class ComplexToStringMixin(models.Model):
    """Mixin that allows to store a complex to string method in the database"""

    COMPUTED_STR_RECOMPUTE_PERIODICALLY: bool = True
    COMPUTED_STR_RECOMPUTE_ON_SAVE: bool = True
    COMPUTED_STR_MAX_LENGTH: int = 512

    class Meta:
        abstract = True

    computed_str = models.CharField(max_length=COMPUTED_STR_MAX_LENGTH, null=True, blank=True, verbose_name=_("Name"))

    def compute_str(self):
        raise NotImplementedError

    def __str__(self):
        if self.computed_str:
            return self.computed_str
        return ""

    def save(self, *args, **kwargs):
        if self.COMPUTED_STR_RECOMPUTE_ON_SAVE or not self.id:
            self.computed_str = self.compute_str()

        if (
            self.computed_str and len(self.computed_str) > self.COMPUTED_STR_MAX_LENGTH
        ):  # ensure computed str does not exceed the provided max length
            self.computed_str = (
                self.computed_str[: self.COMPUTED_STR_MAX_LENGTH - 3] + "..."
            )  # we materialize the fact that we pruned the computed str with ...

        super().save(*args, **kwargs)

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}}"


class LabelKeyMixin:
    @classmethod
    def get_label_key(cls):
        if hasattr(cls, "LABEL_KEY"):
            return cls.LABEL_KEY
        raise AssertionError(
            f"You need to implement the get_label_key method in the class {cls} or remove the LabelKeyMixin from {cls}."
        )


class Status(ChoiceEnum):
    PENDING = _("Pending")
    DENIED = _("Denied")
    APPROVED = _("Approved")

    @classmethod
    def get_color_map(cls):
        colors = [WBColor.YELLOW_LIGHT.value, WBColor.RED_LIGHT.value, WBColor.GREEN_LIGHT.value]
        return [choice for choice in zip(cls, colors)]


class PrimaryMixin(models.Model):
    """
    Ensures that there is only one entry with primary=True
    for models with a foreign key field to a different model and a primary field.
    """

    PRIMARY_ATTR_FIELDS: list = []  # define the list of fields that needs to introduce uniqueness
    primary = models.BooleanField(default=False, verbose_name="Primary")

    @property
    def can_update_primary_field(self) -> bool:
        """
        Check if any of the primary fields have non-None values for updating.

        This property evaluates whether any of the primary fields defined in
        `PRIMARY_ATTR_FIELDS` have non-None values in the current instance.
        It is useful to determine if the primary fields can be updated.

        Returns:
            bool: True if any primary field has a non-None value, False otherwise.
        """
        return any([get_attribute(self, field.split(".")) is not None for field in self.PRIMARY_ATTR_FIELDS])

    def get_related_queryset(self):
        """
        Retrieve a queryset of related objects based on primary attributes.

        This method generates a queryset for related objects of the same model
        by filtering on primary attributes. Primary attributes are defined in
        the class attribute `PRIMARY_ATTR_FIELDS`.

        Returns:
            QuerySet: A Django QuerySet containing related objects.

        """
        return self._meta.model.objects.filter(
            **{field.replace(".", "__"): get_attribute(self, field.split(".")) for field in self.PRIMARY_ATTR_FIELDS}
        )

    def save(self, *args, **kwargs):
        if self.can_update_primary_field:
            related_qs = self.get_related_queryset()
            if self.primary:
                related_qs = related_qs.exclude(id=self.id)  # if self id is None, it will not exclude anything
                related_qs.update(primary=False)
            elif not related_qs.filter(primary=True).exists():
                self.primary = True
        super().save(*args, **kwargs)

    def delete(self, no_deletion=True, **kwargs):
        if self.primary and (qs := self.get_related_queryset().exclude(id=self.id)).exists():
            next_primary = qs.first()
            next_primary.primary = True
            next_primary.save()
            self.primary = False
            self.save()
        if no_deletion:
            for field_name in self.PRIMARY_ATTR_FIELDS:
                with suppress(FieldError, FieldDoesNotExist):
                    if self._meta.get_field(field_name).null:  # ensure the field is nullable
                        setattr(
                            self, field_name, None
                        )  # We don't delete the object, we just un-assign the entry to "hide" it
            self.save()
        else:
            super().delete(**kwargs)

    class Meta:
        abstract = True


class DefaultMixin(models.Model):
    """
    Ensures that there is only one instance with default=True.
    """

    default = models.BooleanField(
        verbose_name=_("Default"),
        default=False,
    )

    def save(self, *args, **kwargs):
        if hasattr(self, "default"):
            if self.default:
                self._meta.model.objects.all().update(default=False)
            elif self._meta.model.objects.all().count() == 0:
                self.default = True
        return super().save(*args, **kwargs)

    class Meta:
        abstract = True


class ActiveObjectManager(OrderedModelManager):
    def get_queryset(self):
        try:
            return super().get_queryset().filter(is_active=True)
        except FieldError:
            return super().get_queryset()


class DeleteToDisableMixin(models.Model):
    AUTOMATICALLY_CLEAN_SOFT_DELETED_OBJECTS = False

    is_active = models.BooleanField(default=True)
    deletion_datetime = models.DateTimeField(blank=True, null=True)

    objects = ActiveObjectManager()
    all_objects = models.Manager()

    def delete(self, using=None, no_deletion=True, **kwargs):
        if no_deletion:
            pre_delete.send(
                sender=self.__class__,
                instance=self,
            )
            self.is_active = False
            self.deletion_datetime = timezone.now()
            self.save()
            post_delete.send(sender=self.__class__, instance=self, using=using)
        else:
            super().delete(using=using, **kwargs)

    class Meta:
        abstract = True


def get_object(model_class: models.Model, object_id: int):
    if issubclass(model_class, DeleteToDisableMixin):
        objects = model_class.all_objects
    else:
        objects = model_class.objects
    return objects.get(id=object_id)


class CalendarItemTypeMixin(models.Model):
    color = ColorField(
        verbose_name=_("Color"),
        default="#000000",
        unique=False,
        blank=False,
        null=False,
    )

    icon = IconField(
        max_length=128,
        unique=False,
        verbose_name=_("Icon"),
        blank=False,
        null=False,
        default=WBIcon.EVENT.value,
    )

    class Meta:
        abstract = True


class ResolvableModelMixin(models.Model):
    class Meta:
        abstract = True

    resolved = models.BooleanField(default=False, verbose_name="Resolved")

    def delete(self, *args, **kwargs):
        assert not self.resolved, "Resolved Entities cannot be deleted."
        super().delete(*args, **kwargs)



import json
import types

from django.core.serializers.json import DjangoJSONEncoder
from rest_framework.renderers import JSONRenderer


class PandasDataCustomEncoder(json.JSONEncoder):
    def default(self, obj):
        # We need to handle generator for option request that sometime contains generator as values
        if isinstance(obj, types.GeneratorType):
            return list(obj)
        return DjangoJSONEncoder().default(obj)


class PandasDataCustomRenderer(JSONRenderer):
    encoder_class = PandasDataCustomEncoder
    strict = False



from typing import Any, Dict, Optional
from uuid import uuid4

from django.contrib.sites.models import Site
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse
from wbcore.models import WBModel


class ShareableLink(WBModel):
    document = models.ForeignKey(to="documents.Document", related_name="shareable_links", on_delete=models.CASCADE)

    valid_until = models.DateTimeField(null=True, blank=True, verbose_name=_("Valid Until"))
    one_time_link = models.BooleanField(default=False, verbose_name=_("One Time Link"))
    manual_invalid = models.BooleanField(default=False)

    uuid = models.UUIDField(default=uuid4, editable=False, unique=True)

    def __str__(self) -> str:
        return f"{self.document}: {self.valid_until}"

    def is_valid(self):
        return (
            not self.manual_invalid
            and (self.valid_until is None or self.valid_until > timezone.now())
            and (not self.one_time_link or self.accesses.count() < 1)
        )

    @property
    def link(self):
        endpoint = reverse("wbcore:documents:download", kwargs={"uuid": self.uuid})
        return "https://%s%s" % (Site.objects.get_current().domain, endpoint)

    def access(self, metadata: Optional[Dict[str, Any]] = None):
        if not metadata:
            metadata = dict()
        ShareableLinkAccess.objects.create(shareable_link=self, metadata=metadata)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:documents:link"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:documents:linkrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{uuid}}"

    class Meta:
        verbose_name = _("Shareable Link")
        verbose_name_plural = _("Shareable Links")


class ShareableLinkAccess(WBModel):
    shareable_link = models.ForeignKey(to="documents.ShareableLink", related_name="accesses", on_delete=models.PROTECT)

    metadata = models.JSONField(default=dict)

    accessed = models.DateTimeField(auto_now_add=True)

    def __str__(self) -> str:
        return f"{self.shareable_link}: {self.accessed}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:documents:linkaccess"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:documents:linkaccessrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{id}}"

    class Meta:
        verbose_name = _("Shareable Link Access")
        verbose_name_plural = _("Shareable Link Accesses")



from dataclasses import dataclass, field
from typing import Iterable, Literal, Optional

from wbcore.metadata.configs.display.formatting import Formatting, FormattingRule


@dataclass
class Tooltip:
    key: str | None = None
    endpoint: str | None = None
    width: str | None = None
    height: str | None = None

    def __post_init__(self):
        if not self.key and not self.endpoint:
            raise ValueError("Either key or endpoint needs to be specified.")

    def serialize(self):
        tooltip = {}
        if self.key:
            tooltip["key"] = self.key

        if self.endpoint:
            tooltip["endpoint"] = self.endpoint

        if self.width:
            tooltip["width"] = self.width

        if self.height:
            tooltip["height"] = self.height

        return tooltip


@dataclass(unsafe_hash=True)
class Field:
    key: str
    label: str
    formatting_rules: Iterable[FormattingRule] = field(default_factory=list)
    width: int | None = None

    tooltip: Tooltip | None = None

    hide: bool | None = None
    pinned: str | None = None

    children: list["Field"] | None = None
    marry_children: bool | None = True
    show: Literal["open"] | Literal["closed"] | None = None
    open_by_default: bool | None = None
    supress_movable: bool = False
    lock_position: bool = False
    resizable: bool = True

    def iterate_leaf_fields(self, aggregated_parent_label: str = ""):
        label = self.label
        if aggregated_parent_label:
            label = aggregated_parent_label + " " + label
        if self.children:
            for children in self.children:
                yield from children.iterate_leaf_fields(label)
        else:
            yield self.key, label

    def __iter__(self):
        yield "key", self.key
        yield "label", self.label
        yield "formatting_rules", [dict(rule) for rule in self.formatting_rules]

        if self.width:
            yield "width", self.width

        if self.hide:
            yield "hide", self.hide

        if self.pinned and self.pinned in ["left", "right"]:
            yield "pinned", self.pinned

        if self.children:
            yield "children", [dict(child) for child in self.children]
            yield "marry_children", self.marry_children is True  # Convert None into False

        if self.show:
            yield "show", self.show

        if self.open_by_default is not None:
            yield "open_by_default", self.open_by_default
        yield "supress_movable", self.supress_movable
        yield "lock_position", self.lock_position

        if self.tooltip:
            yield "tooltip", self.tooltip.serialize()


@dataclass(unsafe_hash=True)
class LegendItem:
    icon: str
    label: str
    value: Optional[str | bool] = None

    def __iter__(self):
        yield "icon", self.icon
        yield "label", self.label

        if self.value is not None:
            yield "value", self.value


@dataclass(unsafe_hash=True)
class Legend:
    items: list[LegendItem]
    label: str | None = None
    key: str | None = None

    def __post_init__(self):
        if self.key:
            assert all(
                [item.value is not None for item in self.items]
            ), "If key is set, all items need to specify a value."

    def __iter__(self):
        if self.label:
            yield "label", self.label

        if self.key:
            yield "key", self.key

        yield "items", [dict(item) for item in self.items]


@dataclass(unsafe_hash=True)
class TreeGroupLevelOption:
    list_endpoint: str
    reorder_endpoint: str | None = None
    reparent_endpoint: str | None = None
    lookup: str = "_group_key"
    filter_key: str = "group_keys"
    filter_whitelist: list[str] = field(default_factory=list)
    filter_blacklist: list[str] = field(default_factory=list)
    clear_filter: bool = (
        False
        # Set to True if preselected filters other than required ones need to be cleared out before fetching the tree group
    )
    filter_depth: int = 1
    parent_field: str = "parent"
    ordering_field: str = "order"

    def __iter__(self):
        endpoints = {"list": self.list_endpoint}
        if self.reorder_endpoint:
            endpoints["reorder"] = self.reorder_endpoint

        if self.reparent_endpoint:
            endpoints["reparent"] = self.reparent_endpoint

        yield "endpoints", endpoints
        yield "lookup", self.lookup
        yield "filter_key", self.filter_key
        yield "filter_whitelist", self.filter_whitelist
        yield "filter_blacklist", self.filter_blacklist
        yield "clear_filter", self.clear_filter
        yield "filter_depth", self.filter_depth
        yield "parent_field", self.parent_field
        yield "ordering_field", self.ordering_field


@dataclass(unsafe_hash=True)
class ListDisplay:
    fields: list[Field | None]
    legends: list[Legend] = field(default_factory=list)
    formatting: list[Formatting] = field(default_factory=list)
    editable: bool = True
    hide_control_bar: bool = False

    ordering_field: str = "order"

    tree: bool = False
    tree_group_pinned: str = "left"
    tree_group_field: str | None = None
    tree_group_label: str | None = None
    tree_group_open_level: int = 0
    tree_group_field_sortable: bool = False

    tree_group_level_options: list[TreeGroupLevelOption] = field(default_factory=list)

    @property
    def flatten_fields(self) -> tuple[str, str]:
        if self.fields:  # fields can be None
            for f in filter(lambda o: o, self.fields):  # field can be None
                yield from f.iterate_leaf_fields()

    def __iter__(self):
        yield "editable", self.editable
        yield "fields", [dict(field) for field in self.fields if field]
        yield "legends", [dict(legend) for legend in self.legends if legend]
        yield "formatting", [dict(formatting) for formatting in self.formatting if formatting]
        yield "hide_control_bar", self.hide_control_bar
        yield "ordering_field", self.ordering_field
        if self.tree:
            yield "tree", self.tree
            tree_group = {
                "pinned": self.tree_group_pinned,
                "field": self.tree_group_field if self.tree_group_field else self.fields[0].key,
                "label": self.tree_group_label if self.tree_group_label else self.fields[0].key,
                "open_level": self.tree_group_open_level,
                "field_sortable": self.tree_group_field_sortable,
            }
            tree_group["level_options"] = [dict(level_option) for level_option in self.tree_group_level_options]
            yield "tree_group", tree_group


@dataclass(unsafe_hash=True)
class Calendar:
    title: str
    period: str
    all_day: bool
    color: str
    icon: str
    entities: list
    endpoint: str

    def __iter__(self):
        yield "calendar", {
            "title": self.title,
            "period": self.period,
            "all_day": self.all_day,
            "color": self.color,
            "icon": self.icon,
            "entities": self.entities,
            "endpoint": self.endpoint,
        }



from django.utils.translation import gettext as _
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class ShareableLinkModelButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.ActionButton(
                method=RequestType.PATCH,
                key="manually_invalidate",
                label=_("Invalidate"),
                icon=WBIcon.UNLINK.icon,
                description_fields=_("<p> Are you sure you want to invalidate this link? </p>"),
                title=_("Invalidate"),
                action_label=_("Invalidation"),
            )
        }



import pytest
from django.contrib.contenttypes.models import ContentType
from rest_framework.test import APIRequestFactory
from wbcore.contrib.workflow.models import Step
from wbcore.contrib.workflow.viewsets import (
    EmailStepModelViewSet,
    StepModelViewSet,
    TransitionModelViewSet,
    WorkflowModelViewSet,
)
from wbcore.test.utils import get_or_create_superuser


@pytest.mark.django_db
class TestWorkflow:
    api_factory = APIRequestFactory()

    def test_filter_attached_model(self, workflow_factory):
        workflow_factory()
        workflow_factory(model=ContentType.objects.first())
        mvs = WorkflowModelViewSet()
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        qs = mvs.get_queryset()
        assert mvs.filterset_class(request=request).filter_attached_model(qs, "", None) == qs
        assert mvs.filterset_class(request=request).filter_attached_model(qs, "", "person").count() == 1
        assert not mvs.filterset_class(request=request).filter_attached_model(qs, "", "company").exists()

    def test_filter_by_data(self, workflow_factory, data_factory):
        workflow_factory()
        data = data_factory()
        mvs = WorkflowModelViewSet()
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        qs = mvs.get_queryset()
        assert mvs.filterset_class(request=request).filter_by_data(qs, "", None) == qs
        assert mvs.filterset_class(request=request).filter_by_data(qs, "", data).count() == 1
        assert mvs.filterset_class(request=request).filter_by_data(qs, "", data).first() == data.workflow


@pytest.mark.django_db
class TestStep:
    api_factory = APIRequestFactory()

    def test_filter_by_transition(self, transition_factory, random_child_step_factory):
        transition = transition_factory()
        random_child_step_factory()
        mvs = StepModelViewSet(kwargs={})
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        qs = mvs.get_queryset()
        assert mvs.get_filterset_class(request=request)().filter_by_transition(qs, "", None) == qs
        assert set(mvs.get_filterset_class(request=request)().filter_by_transition(qs, "", transition)) == set(
            Step.objects.filter(id__in=[transition.from_step.pk, transition.to_step.pk])
        )

    def test_email_filter_template_name(self, email_step_factory):
        step = email_step_factory()
        email_step_factory()
        mvs = EmailStepModelViewSet(kwargs={})
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        qs = mvs.get_queryset()
        assert mvs.get_filterset_class(request=request)().filter_template_name(qs, "", None) == qs
        assert mvs.get_filterset_class(request=request)().filter_template_name(qs, "", step.template.name).count() == 1
        assert (
            mvs.get_filterset_class(request=request)().filter_template_name(qs, "", step.template.name).first() == step
        )


@pytest.mark.django_db
class TestTransition:
    api_factory = APIRequestFactory()

    def test_filter_by_condition(self, condition_factory, transition_factory):
        transition_factory()
        condition = condition_factory()
        mvs = TransitionModelViewSet(kwargs={})
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        qs = mvs.get_queryset()
        assert mvs.filterset_class(request=request).filter_by_condition(qs, "", None) == qs
        assert mvs.filterset_class(request=request).filter_by_condition(qs, "", condition).count() == 1
        assert (
            mvs.filterset_class(request=request).filter_by_condition(qs, "", condition).first() == condition.transition
        )



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class ReleaseNoteDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="module", label=_("Module")),
                dp.Field(key="version", label=_("Version")),
                dp.Field(key="release_date", label=_("Release Date")),
                dp.Field(key="summary", label=_("Summary")),
            ],
            formatting=[
                dp.Formatting(
                    column="user_read",
                    formatting_rules=[
                        dp.FormattingRule(icon=WBIcon.VIEW.icon, condition=("==", True)),
                    ],
                ),
            ],
            legends=[
                dp.Legend(
                    key="read_unread",
                    items=[
                        dp.LegendItem(
                            icon=WBIcon.VIEW.icon,
                            label=_("Read"),
                            value="read",
                        ),
                        dp.LegendItem(
                            icon=WBIcon.IGNORE.icon,
                            label=_("Unread"),
                            value="unread",
                        ),
                    ],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["module", "version", "release_date"],
                [repeat_field(3, "summary")],
                [repeat_field(3, "notes")],
            ]
        )



from wbcore.enums import WidgetType  # TODO: Rename to WindowType

from .base import WBCoreViewConfig


class WindowTypeViewConfig(WBCoreViewConfig):
    metadata_key = "type"
    config_class_attribute = "window_type_config_class"

    def get_metadata(self) -> str:
        if window_type := getattr(self.view, "WIDGET_TYPE", None):
            return window_type
        return WidgetType.INSTANCE.value if self.instance else WidgetType.LIST.value



from django.db.models import IntegerChoices
from import_export.admin import ImportExportMixin


class ExportFormat(IntegerChoices):
    CSV = 0, "CSV"
    XLSX = 1, "XLSX"
    TSV = 2, "TSV"
    JSON = 3, "JSON"
    YAML = 4, "YAML"


def get_django_import_export_format(format: int):
    format_map = {f.__name__: f for f in ImportExportMixin().get_export_formats()}
    return format_map[ExportFormat(format).name]



import pytest
from django.test import Client, TestCase
from django.urls import reverse
from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.example_app.factories import LeagueFactory
from wbcore.contrib.example_app.models import League
from wbcore.contrib.example_app.serializers import LeagueModelSerializer
from wbcore.contrib.example_app.tests.test_viewsets.test_utils_viewsets import (
    find_instances_in_response,
    get_create_view,
    get_delete_view,
    get_detail_view,
    get_partial_view,
    get_update_view,
)
from wbcore.contrib.example_app.viewsets import LeagueModelViewSet


@pytest.mark.django_db
class TestLeagueModelViewSet(TestCase):
    def setUp(self):
        self.user = SuperUserFactory()
        self.client = Client()
        self.client.force_login(user=self.user)
        self.list_url = reverse("example_app:league-list")
        self.detail_url_str = "example_app:league-detail"

    def test_list_view(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

    def test_create_view(self):
        league = LeagueFactory()
        response = get_create_view(self.client, league, self.user, self.list_url, LeagueModelViewSet)
        self.assertEqual(response.status_code, 201)
        self.assertTrue(League.objects.filter(name=league.name).exists())

    def test_detail_view(self):
        instance = LeagueFactory()
        response = get_detail_view(self.client, instance.pk, self.detail_url_str)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_update_view(self):
        instance = LeagueFactory()
        instance.name = "Updated Instance"
        response = get_update_view(self.client, instance, LeagueModelSerializer, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_partial_update_view(self):
        instance = LeagueFactory()
        response = get_partial_view(self.client, instance.id, {"name": "Updated Instance"}, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_delete_view(self):
        instance = LeagueFactory()
        response = get_delete_view(self.client, self.detail_url_str, instance.pk)
        self.assertEqual(response.status_code, 204)
        self.assertFalse(League.objects.filter(pk=instance.pk).exists())

    def test_league_sport(self):
        league_a = LeagueFactory()
        league_b = LeagueFactory(sport=league_a.sport)
        league_c = LeagueFactory()
        expected_number_of_league = League.objects.filter(sport=league_a.sport).count()
        league_sport_url = reverse("example_app:league-sport-list", args=[league_a.sport.id])
        response = self.client.get(league_sport_url)
        league_a_found, league_b_found, league_c_found = find_instances_in_response(
            [league_a, league_b, league_c], response
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.data["count"],
            expected_number_of_league,
            f"The answer should contain {expected_number_of_league} leagues",
        )
        self.assertTrue(league_a_found, "Player A was not found in Response.")
        self.assertTrue(league_b_found, "Player B was not found in Response.")
        self.assertFalse(league_c_found, "Player C was found in Response, but should not be found.")



from typing import Optional

from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.lookups import Unaccent
from django.contrib.postgres.search import SearchVector, SearchVectorField
from django.db import models
from django.db.models.expressions import Value
from django.db.models.functions import Lower
from django.db.models.query import QuerySet
from mptt.managers import TreeManager
from mptt.models import MPTTModel, TreeForeignKey
from text_unidecode import unidecode
from timezone_field import TimeZoneField
from wbcore.models import WBModel


class LeveledGeographyManager(TreeManager):
    def get_by_natural_key(self, code_2):
        if self.level == 1:
            return self.get(code_2=code_2)

    def get_by_name(self, name, **extra_filter_args):
        qs = self
        if extra_filter_args:
            qs = qs.filter(**extra_filter_args)
        return (
            qs.annotate(name_unaccent=Unaccent(Lower("name")))
            .filter(models.Q(name_unaccent=unidecode(name).lower()) | models.Q(alternative_names__contains=[name]))
            .order_by("ranking")
            .first()
        )

    def __init__(self, level: Optional[int] = None, *args, **kwargs):
        self.level = level
        super().__init__(*args, **kwargs)

    def get_queryset(self) -> QuerySet:
        queryset = super().get_queryset()
        if level := self.level:
            return queryset.filter(level=level)
        return queryset


class Geography(WBModel, MPTTModel):
    class Level(models.IntegerChoices):
        CONTINENT = 0, "Continent"
        COUNTRY = 1, "Country"
        STATE = 2, "State"
        CITY = 3, "City"

    name = models.CharField(max_length=255)
    short_name = models.CharField(max_length=255)
    alternative_names = ArrayField(
        base_field=models.CharField(max_length=255),
        default=list,
        blank=True,
        verbose_name="Alternative Names",
        help_text="Alternative Names",
    )
    parent = TreeForeignKey("self", on_delete=models.CASCADE, null=True, blank=True, related_name="children")
    representation = models.CharField(max_length=255, default="", blank=True)

    search_vector = SearchVectorField(null=True)
    trigram_search_vector = models.CharField(max_length=1024, null=True, blank=True)

    code_2 = models.CharField(max_length=2, verbose_name="2 Character Alphanumeric Code", null=True, blank=True)
    code_3 = models.CharField(max_length=3, verbose_name="3 Character Alphanumeric Code", null=True, blank=True)
    code = models.CharField(max_length=16, verbose_name="Code", null=True, blank=True)

    population = models.PositiveIntegerField(null=True, blank=True)
    ranking = models.PositiveIntegerField(default=1)
    time_zone = TimeZoneField(choices_display="WITH_GMT_OFFSET", null=True, blank=True)

    objects = LeveledGeographyManager()
    continents = LeveledGeographyManager(level=0)
    countries = LeveledGeographyManager(level=1)
    states = LeveledGeographyManager(level=2)
    cities = LeveledGeographyManager(level=3)

    def lookup_descendants(self, name: str, level: Optional[int] = None) -> models.Model | None:
        qs = self.get_descendants()
        if level is not None:
            qs = qs.filter(level=level)

        qs = qs.annotate(name_unaccent=Unaccent(Lower("name"))).filter(
            models.Q(name_unaccent=unidecode(name).lower()) | models.Q(alternative_names__contains=[name])
        )
        if qs.count() == 1:
            return qs.first()

    class Meta:
        verbose_name_plural = "Geographies"
        indexes = [
            GinIndex(fields=["search_vector"], name="geography_sv_gin_idx"),  # type: ignore
            GinIndex(fields=["trigram_search_vector"], opclasses=["gin_trgm_ops"], name="geography_trigram_sv_gin_idx"),  # type: ignore
        ]
        constraints = [
            models.UniqueConstraint(
                name="%(app_label)s_%(class)s_level_code2__uniq",
                fields=("level", "code_2"),
                condition=models.Q(level=1),
            ),
            models.UniqueConstraint(
                name="%(app_label)s_%(class)s_level_code3__uniq",
                fields=("level", "code_3"),
                condition=models.Q(level=1),
            ),
        ]

    def update_search_vectors(self):
        alternative_names = [Value(name) for name in self.alternative_names]
        self.search_vector = SearchVector(Value(self.name), *alternative_names)
        self.trigram_search_vector = f"{self.name} {' '.join(self.alternative_names)}".strip()

    def save(self, *args, **kwargs):
        self.update_search_vectors()
        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return f"{self.name}"

    @classmethod
    def dict_to_model(cls, country_data, level=Level.COUNTRY):
        qs = Geography.objects.filter(level=level)
        if isinstance(country_data, int):
            return qs.filter(id=country_data).first()
        elif code := country_data.get("code", None):
            return qs.filter(code_3=code).first()
        elif code_alpha2 := country_data.get("code_alpha2", None):
            return qs.filter(code_2=code_alpha2).first()

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:geography:geography"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:geography:geographyrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{representation}}"



from unittest.mock import patch

import pytest
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.workflow.models import Condition


@pytest.mark.django_db
class TestDecisionStep:
    def test_get_first_valid_transition(
        self, process_step_factory, decision_step_factory, transition_factory, condition_factory
    ):
        attached_instance = PersonFactory()
        step = decision_step_factory()
        process_step = process_step_factory(process__instance=attached_instance, step=step)
        valid_transition1 = transition_factory(from_step=step)
        valid_transition2 = transition_factory(from_step=step)
        invalid_transition = transition_factory(from_step=step)
        transition_factory()
        condition_factory(
            transition=invalid_transition,
            attribute_name="first_name",
            expected_value=attached_instance.first_name,
            operator=Condition.Operator.EQ,
            negate_operator=True,
        )
        assert step.get_first_valid_transition(process_step) in [valid_transition1, valid_transition2]

    def test_get_first_valid_transition_no_transitions(
        self, process_step_factory, decision_step_factory, transition_factory, condition_factory
    ):
        attached_instance = PersonFactory()
        step = decision_step_factory()
        process_step = process_step_factory(process__instance=attached_instance, step=step)
        invalid_transition = transition_factory(from_step=step)
        transition_factory()
        condition_factory(
            transition=invalid_transition,
            attribute_name="first_name",
            expected_value=attached_instance.first_name,
            operator=Condition.Operator.EQ,
            negate_operator=True,
        )
        assert step.get_first_valid_transition(process_step) is None

    @patch("wbcore.contrib.workflow.models.step.DecisionStep.get_first_valid_transition")
    @patch("wbcore.contrib.workflow.models.step.Step.start_next_step")
    def test_run(
        self,
        mock_next,
        mock_transition,
        process_step_factory,
        transition_factory,
        decision_step_factory,
    ):
        transition = transition_factory()
        mock_transition.return_value = transition
        step = decision_step_factory()
        process_step = process_step_factory(step=step)
        step.run(process_step)
        assert mock_next.call_args.args == (process_step, transition)

    @patch("wbcore.contrib.workflow.models.step.Step.set_failed")
    @patch("wbcore.contrib.workflow.models.step.activate_step.delay")
    @patch("wbcore.contrib.workflow.models.step.DecisionStep.get_first_valid_transition")
    def test_run_failed(
        self,
        mock_transition,
        mock_activate,
        mock_failed,
        process_step_factory,
        decision_step_factory,
    ):
        mock_transition.return_value = None
        step = decision_step_factory()
        process_step = process_step_factory(step=step)
        step.run(process_step)
        assert mock_failed.call_args.args[0] == process_step
        assert not mock_activate.called



from contextlib import suppress

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _
from wbcore.utils.models import PrimaryMixin


class DocumentModelRelationship(PrimaryMixin):
    PRIMARY_ATTR_FIELDS = ["content_type", "object_id"]
    document = models.ForeignKey(to="documents.Document", related_name="relationships", on_delete=models.CASCADE)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")
    content_object_repr = models.CharField(max_length=512, null=True, blank=True)

    def __str__(self) -> str:
        return f"{self.document.name} -> {self.content_object}"

    def delete(self, no_deletion=True, **kwargs):
        super().delete(no_deletion=False, **kwargs)

    def save(self, *args, **kwargs):
        with suppress(AttributeError):
            self.content_object_repr = str(self.content_object)
        return super().save(*args, **kwargs)

    class Meta:
        verbose_name = _("Document Model Relationship")
        verbose_name_plural = _("Document Model Relationships")

        indexes = [models.Index(fields=["content_type", "object_id"])]
        constraints = [
            models.UniqueConstraint(
                fields=["content_type", "object_id"],
                name="unique_primary_relationship",
                condition=models.Q(primary=True),
            ),
            models.UniqueConstraint(fields=["document", "content_type", "object_id"], name="unique_relationship"),
        ]

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcore:documents:documentmodelrelationship"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    # @classmethod
    # def get_representation_endpoint(cls):
    #     return "wbcore:documents:documentmodelrelationshiprepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{content_type}}: {{content_object_repr}}"



import logging

from django.conf import settings
from django.contrib.auth import user_logged_in
from django.contrib.auth.password_validation import validate_password
from django.forms import ValidationError
from django.utils import timezone
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from jwt import decode
from rest_framework import exceptions, serializers
from rest_framework.reverse import reverse
from rest_framework_simplejwt.serializers import (
    TokenObtainPairSerializer,
    TokenRefreshSerializer,
)
from rest_framework_simplejwt.tokens import RefreshToken
from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.models import Person
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer

from ..models import Group, Permission, User
from ..models.users_activities import refresh_user_activity

logger = logging.getLogger()


class ChangePasswordSerializer(wb_serializers.Serializer):
    old_password = wb_serializers.CharField(label=gettext_lazy("Old password"), required=True, secure=True)
    new_password = wb_serializers.CharField(label=gettext_lazy("New Password"), required=True, secure=True)
    confirm_password = wb_serializers.CharField(label=gettext_lazy("Confirm new Password"), required=True, secure=True)

    def validate(self, data):
        if not (request := self.context.get("request")):
            raise ValueError(_("No request found"))
        user = request.user
        # user = data.get("user")
        current_password = data.get("old_password")
        if not user.check_password(current_password):
            raise serializers.ValidationError({"old_password": _("Wrong password")})

        new_password = data.get("new_password")
        confirm_password = data.get("confirm_password")
        if new_password != confirm_password:
            raise serializers.ValidationError(
                {"new_password": _("Password don't match"), "confirm_password": _("Password don't match")}
            )

        if user.check_password(new_password):
            raise serializers.ValidationError(
                {
                    "new_password": _("New and old password cannot be the same"),
                    "confirm_password": _("New and old password cannot be the same"),
                }
            )

        if user.check_password(current_password):
            try:
                validate_password(new_password)
                return super().validate(data)
            except serializers.ValidationError as e:
                raise serializers.ValidationError({"new_password": e, "confirm_password": e})

    class Meta:
        fields = ("current_password", "new_password", "confirm_password")


class UserRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:authentication:user-detail")
    computed_str = wb_serializers.SerializerMethodField()

    def get_computed_str(self, obj):
        if obj.profile:
            return obj.profile.computed_str
        return f"{obj.username} ({obj.email})"

    class Meta:
        model = User
        fields = ("id", "username", "computed_str", "email", "_detail")


class PermissionRepresentationField(serializers.RelatedField):
    def to_representation(self, value):
        return "{0.content_type.app_label}.{0.codename}".format(value)


class GroupRepresentationSerializer(wb_serializers.RepresentationSerializer):
    # _detail = wb_serializers.HyperlinkField(reverse_name='wbcore:authentication:group-detail')
    value_key = "id"
    label_key = "name"

    class Meta:
        model = Group
        fields = ("id", "name")


class PermissionRepresentationSerializer(wb_serializers.RepresentationSerializer):
    # _detail = serializers.SerializerMethodField()

    endpoint = "wbcore:authentication:permissionrepresentation-list"
    value_key = "id"
    label_key = "{{name}}"

    # _detail = wb_serializers.HyperlinkField(reverse_name='wbcore:authentication:permission-detail')

    class Meta:
        model = Permission
        fields = ("id", "name")


class GroupModelSerializer(wb_serializers.ModelSerializer):
    _permissions = PermissionRepresentationSerializer(many=True, read_only=True, source="permissions")

    class Meta:
        model = Group
        fields = (
            "id",
            "name",
            "permissions",
            "_permissions",
        )


class PermissionModelSerializer(wb_serializers.ModelSerializer):
    content_type_repr = wb_serializers.SerializerMethodField()

    class Meta:
        model = Permission
        fields = ("id", "name", "content_type", "content_type_repr", "codename", "group_set")

    def get_content_type_repr(self, obj):
        return str(obj.content_type.app_label)


class UserModelSerializer(wb_serializers.ModelSerializer):
    _groups = GroupRepresentationSerializer(source="groups", many=True)
    _user_permissions = PermissionRepresentationSerializer(source="user_permissions", many=True)
    _profile = PersonRepresentationSerializer(many=False, required=False, source="profile")

    is_active = wb_serializers.BooleanField(default=True, label=gettext_lazy("Is Active"))
    is_superuser = wb_serializers.BooleanField(default=False)
    last_connection = wb_serializers.DateTimeField(read_only=True, label=gettext_lazy("Last Connection"))
    email = wb_serializers.CharField(required=False, label=gettext_lazy("E-Mail"))
    username = wb_serializers.CharField(required=False, label=gettext_lazy("Username"))
    profile = wb_serializers.PrimaryKeyRelatedField(queryset=Person.objects.all(), label=gettext_lazy("Profile"))

    @wb_serializers.register_resource()
    def reset_password(self, instance, request, user):
        res = {
            "reset_password": reverse(
                "wbcore:authentication:user-reset-password", args=[instance.id], request=request
            ),
            "user_permissions": reverse(
                "wbcore:authentication:user-permissions-list", args=[instance.id], request=request
            ),
        }
        if instance.profile:
            res["profile"] = reverse("wbcore:directory:person-detail", args=[instance.profile.id], request=request)
        return res

    @wb_serializers.register_resource()
    def user_activity(self, instance, request, user):
        additional_resources = dict()
        additional_resources["user_activity"] = reverse(
            "wbcore:authentication:user-useractivity-list", args=[instance.id], request=request
        )
        additional_resources["user_activity_chart"] = reverse(
            "wbcore:authentication:user-useractivitychart-list", args=[instance.id], request=request
        )
        return additional_resources

    def get_all_permissions(self, obj):
        permissions = list(obj.user_permissions.all().values_list("id", flat=True))
        for group in obj.groups.all():
            group_permissions = group.permissions.all().values_list("id", flat=True)
            permissions.extend(list(group_permissions))
        return list(set(permissions))

    def get_all_permissions_repr(self, obj):
        permissions = list()
        permissions_dict = list(obj.user_permissions.all().values_list("content_type__app_label", "codename"))
        permissions.extend(["{0[0]}.{0[1]}".format(_permission) for _permission in permissions_dict])
        for group in obj.groups.all():
            group_permissions = group.permissions.all().values_list("content_type__app_label", "codename")
            permissions.extend(["{0[0]}.{0[1]}".format(_permission) for _permission in group_permissions])
        return list(set(permissions))

    @classmethod
    def prefetch_related(cls, queryset, request):
        queryset = queryset.prefetch_related("groups")
        queryset = queryset.prefetch_related("user_permissions")
        return queryset

    def validate(self, data):
        obj = self.instance

        username = data.pop("username", obj.username if obj else None)
        email = data.pop("email", obj.email if obj else None)
        errors = {}
        if profile := data.pop("profile", obj.profile if obj else None):
            if not username:
                username = User.generate_username(profile.first_name, profile.last_name)
            if not email and (primary_emails := profile.emails.filter(primary=True).first()):
                email = primary_emails.address

        if not email:
            errors["email"] = _("You need to assign a profile or add a valid email address")
        if not username:
            errors["username"] = _("You need to assign a profile or add a valid last name")

        if len(errors.keys()) > 0:
            raise serializers.ValidationError(errors)
        if (user_account := getattr(profile, "user_account", None)) and (user_account != self.instance):
            errors["profile"] = f"An account already exists for this profile under the email {user_account.email}"
        data["profile"] = profile

        data["username"] = username
        data["email"] = email
        if errors:
            raise ValidationError(errors)
        return super().validate(data)

    class Meta:
        model = User
        fields = (
            "id",
            "username",
            "email",
            "date_joined",
            "last_connection",
            "user_permissions",
            "_user_permissions",
            "groups",
            "_groups",
            # 'all_permissions',
            # 'all_permissions_repr',
            "profile",
            "_profile",
            "is_staff",
            "is_active",
            "is_superuser",
            # 'change_password',
            # 'reset_password',
            # 'main_manager_of',
            "_additional_resources",
        )


class UserProfileModelSerializer(wb_serializers.ModelSerializer):
    username = wb_serializers.CharField()
    email = wb_serializers.ReadOnlyField()
    profile = wb_serializers.PrimaryKeyRelatedField(read_only=True)
    _profile = PersonRepresentationSerializer(many=False, read_only=True, source="profile")
    generic_auth_token_key = wb_serializers.SerializerMethodField(label="Generic Token Key")

    calendar_subscription_link = wb_serializers.SerializerMethodField()
    profile_image = wb_serializers.ImageField(source="profile.profile_image", read_only=True)
    first_name = wb_serializers.CharField(source="profile.first_name", read_only=True)
    prefix = wb_serializers.CharField(source="profile.prefix", read_only=True)
    last_name = wb_serializers.CharField(source="profile.last_name", read_only=True)
    birthday = wb_serializers.DateField(source="profile.birthday", read_only=True)

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        resources = dict()
        if user.id == instance.id:
            resources["change_password"] = reverse(
                "wbcore:authentication:userprofile-change-password", args=[instance.id], request=request
            )
            resources["reset_settings"] = reverse(
                "wbcore:authentication:userprofile-reset-settings", args=[instance.id], request=request
            )
            if user.profile.is_internal:
                resources["see_profile"] = reverse(
                    "wbcore:directory:person-detail", args=[instance.profile.id], request=request
                )
        return resources

    def get_calendar_subscription_link(self, obj):
        if (key := obj.generic_auth_token_key) and (request := self.context.get("request")):
            full_url = request.build_absolute_uri(reverse("wbcore:agenda:get_ics"))
            return f"{full_url}?token={key}"
        return "N/A"

    def get_generic_auth_token_key(self, obj):
        return obj.generic_auth_token_key

    def update(self, instance, validated_data):
        profile = validated_data.get("profile", None)
        if profile:
            instance.profile.profile_image = profile.get("profile_image", instance.profile.profile_image)
            instance.profile.first_name = profile.get("first_name", instance.profile.first_name)
            instance.profile.prefix = profile.get("prefix", instance.profile.prefix)
            instance.profile.last_name = profile.get("last_name", instance.profile.last_name)
            instance.profile.birthday = profile.get("birthday", instance.profile.birthday)
            instance.profile.save()
        return instance

    class Meta:
        model = User
        fields = (
            "id",
            "profile_image",
            "first_name",
            "last_name",
            "prefix",
            "birthday",
            "username",
            "email",
            "_profile",
            "profile",
            "generic_auth_token_key",
            "calendar_subscription_link",
            "_additional_resources",
        )


class MyTokenObtainPairSerializer(TokenObtainPairSerializer):
    def validate(self, attrs):
        authentication_status = True
        try:
            data = super().validate(attrs)
            refresh = RefreshToken(data["refresh"])
            user = User.objects.filter(email=attrs[self.username_field]).first()
            jti = refresh.payload.get("jti", None)
            if user and jti:
                user_logged_in.send(
                    sender=MyTokenObtainPairSerializer.__class__,
                    request=self.context.get("request"),
                    user=user,
                    authentication_status=authentication_status,
                    jti=jti,
                )
        except exceptions.AuthenticationFailed:
            raise exceptions.AuthenticationFailed(
                self.error_messages["no_active_account"],
                "no_active_account",
            )
        return data


class MyTokenRefreshSerializer(TokenRefreshSerializer):
    def validate(self, attrs):
        data = super().validate(attrs)
        decoded_data = decode(attrs["refresh"], settings.SECRET_KEY, algorithms=["HS256"])
        user_id = decoded_data.get("user_id", None)
        jti = decoded_data.get("jti", None)
        if user_id and jti:
            refresh_user_activity.delay(user_id, timezone.now(), jti)
        return data



from slugify import slugify
from wbcore.metadata.metadata import WBCoreMetadata


class PandasMetadata(WBCoreMetadata):
    def determine_metadata(self, request, view):
        metadata = super().determine_metadata(request, view)
        metadata["identifier"] += f"-{slugify(view.__class__.__name__)}"
        metadata["fields"] = view.get_pandas_fields(request).to_dict()
        # if ordering fields is not included in the metadata (e.g. dev didn't explicitly specify a list), then we inject the pandas fields by default
        # it is safe to do that because for pandas table, the ordering happens on AgGrid
        if not metadata["ordering_fields"]:
            metadata["ordering_fields"] = {key: key for key in metadata["fields"].keys()}
        return metadata



from unittest import TestCase

import pytest
from django.core.exceptions import ValidationError as DjangoValidationError
from rest_framework.exceptions import ValidationError
from rest_framework.test import APIRequestFactory
from schwifty import IBAN
from wbcore.contrib.directory.models import ClientManagerRelationship as CMR
from wbcore.contrib.directory.serializers import (
    AddressContactSerializer,
    BankingContactSerializer,
    EmailContactSerializer,
    RelationshipTypeRepresentationSerializer,
    TelephoneContactSerializer,
    WebsiteContactSerializer,
)

from ..factories import (
    AddressContactFactory,
    BankingContactFactory,
    EmailContactFactory,
    EntryFactory,
    PersonFactory,
    TelephoneContactFactory,
    WebsiteContactFactory,
)
from ..viewsets import ClientManagerViewSet, RelationshipModelViewSet


@pytest.mark.django_db
class TestRelationshipTypeSerializers:
    def test_serialize_model(self, relationship_type_factory):
        APIRequestFactory().get("")
        type = relationship_type_factory(title="Test Type")
        serializer = RelationshipTypeRepresentationSerializer(type)

        assert serializer.data
        assert serializer.data["title"] == "Test Type"


@pytest.mark.django_db
class TestRelationshipSerializers:
    def test_serialize_model(self, relationship_factory):
        relationship_factory()
        mvs = RelationshipModelViewSet()
        qs = mvs.get_serializer_class().Meta.model.objects.all()
        qs = mvs.serializer_class()
        assert qs.data


@pytest.mark.django_db
class TestClientManagerSerializers(TestCase):
    def test_create_existing_request(self):
        """
        Validate method should fail when given data for an already existing client manager relationship.
        """
        manager = PersonFactory()
        client = PersonFactory(relationship_managers=[manager])
        relationship = CMR.objects.filter(relationship_manager=manager, client=client)
        relationship.update(primary=True)

        mvs = ClientManagerViewSet()
        serializer = mvs.serializer_class()
        data = {
            "client": client,
            "relationship_manager": manager,
            "primary": True,
        }
        with self.assertRaises(ValidationError):
            serializer.validate(data)

    def test_create_new_request(self):
        """
        Validate method should not fail when given data for a new relationship.
        """
        manager = PersonFactory()
        client = PersonFactory()
        mvs = ClientManagerViewSet()
        serializer = mvs.serializer_class()
        data = {
            "client": client,
            "relationship_manager": manager,
        }
        try:
            validated_data = serializer.validate(data)
        except DjangoValidationError:
            self.fail("New client manager relationship threw error in validation method!")
        else:
            self.assertEqual(validated_data, data)

    def test_self_relationship_managers(self):
        """
        Validate method should fail when client and manager are the same person.
        """
        person = PersonFactory()
        mvs = ClientManagerViewSet()
        serializer = mvs.serializer_class()
        data = {
            "client": person,
            "relationship_manager": person,
        }
        with self.assertRaises(ValidationError):
            serializer.validate(data)

    def test_degrade_primary_manager(self):
        """
        Validate method should fail when trying to make primary manager non-primary.
        """
        manager = PersonFactory()
        client = PersonFactory(relationship_managers=[manager])

        mvs = ClientManagerViewSet()
        serializer = mvs.serializer_class()
        data = {
            "client": client,
            "relationship_manager": manager,
            "primary": False,
        }
        with self.assertRaises(ValidationError):
            serializer.validate(data)


@pytest.mark.django_db
class TestContactSerializers(TestCase):
    def test_create_email_duplicates(self):
        """
        Validate method should fail when given data for an already existing email contact.
        """
        entry = EntryFactory()
        email = EmailContactFactory(entry=entry)

        serializer = EmailContactSerializer()
        data = {
            "entry": entry,
            "address": email.address,
        }
        with self.assertRaises(ValidationError):
            serializer.validate(data)

    def test_create_banking_duplicates(self):
        """
        Validate method should fail when given data for an already existing banking contact.
        """
        entry = EntryFactory()
        BankingContactFactory(entry=entry, iban=IBAN("AD1400080001001234567890").formatted)

        serializer = BankingContactSerializer()
        data = {
            "entry": entry,
            "iban": IBAN("AD1400080001001234567890").formatted,
        }
        with self.assertRaises(ValidationError):
            serializer.validate(data)

    def test_create_telephone_duplicates(self):
        """
        Validate method should fail when given data for an already existing telephone contact.
        """
        entry = EntryFactory()
        phone = TelephoneContactFactory(entry=entry)

        serializer = TelephoneContactSerializer()
        data = {
            "entry": entry,
            "number": phone.number,
        }
        with self.assertRaises(ValidationError):
            serializer.validate(data)

    def test_create_website_duplicates(self):
        """
        Validate method should fail when given data for an already existing website contact.
        """
        entry = EntryFactory()
        website = WebsiteContactFactory(entry=entry)

        serializer = WebsiteContactSerializer()
        data = {
            "entry": entry,
            "url": website.url,
        }
        with self.assertRaises(ValidationError):
            serializer.validate(data)

    def test_create_address_duplicates(self):
        """
        Validate method should fail when given data for an already existing address contact.
        """
        entry = EntryFactory()
        address = AddressContactFactory(entry=entry)

        serializer = AddressContactSerializer()
        data = {
            "entry": entry,
            "street": address.street,
            "street_additional": address.street_additional,
            "zip": address.zip,
            "geography_city": address.geography_city,
        }
        with self.assertRaises(ValidationError):
            serializer.validate(data)



import datetime
import json
from functools import partial
from typing import Any, Dict

import factory
from django.db import models
from django.forms.models import model_to_dict
from factory import Factory
from factory.base import StubObject
from rest_framework.request import Request
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.authentication.models import Token

from .signals import (
    custom_update_data_from_factory,
    custom_update_kwargs,
    get_custom_serializer,
)


def get_all_subclasses(klass):
    for subclass in klass.__subclasses__():
        yield subclass
        yield from get_all_subclasses(subclass)


def all_subclasses(cls):
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in all_subclasses(c)])


def get_model_factory(model):
    # mf = [cls for cls in factory.django.DjangoModelFactory.__subclasses__() if cls._meta.model == model]
    mfs = [cls for cls in all_subclasses(factory.django.DjangoModelFactory) if cls._meta.model == model]
    if is_empty(mfs) or model is None:
        return None
    else:
        mf = [mf for mf in mfs if (hasattr(mf, "__name__") and mf.__name__ == model.__name__ + "Factory")]
        if not is_empty(mf):
            return mf[0]
        else:
            return mfs[0]


def get_data_from_factory(instance, viewset, delete=False, update=False, superuser=None, factory=None):
    """
    Our goal here is to get the serializer dynamically based on the viewset, use this serializer to generate data for the post and update test.
    """
    obj_factory = instance
    if delete and factory:
        instance = factory()
    kwargs = getattr(viewset, "kwargs", {})
    kwargs["pk"] = instance.pk
    viewset.kwargs = kwargs
    mvs = viewset()

    request = APIRequestFactory().get("")
    if superuser:
        request.user = superuser
        request.parser_context = {"view": mvs}
        _request = request
    else:
        _request = Request(request)
    mvs.request = request
    """
        TODO
        request = mvs.initialize_request(request)
        object has no attribute 'action_map'
        action_map is filled according to the parameter given to as_view
        Need to find better approch
    """

    if delete:
        mvs.action = "create"
    elif update:
        mvs.action = "update"
    else:
        mvs.action = "list"
    my_serializer = mvs.get_serializer_class()
    if custom_serializer := get_custom_serializer.send(viewset):
        my_serializer = custom_serializer[0][1]
    my_model = my_serializer.Meta.model

    # fields_models = [m.name for m in my_model._meta.get_fields()]
    dict_fields_models = {}
    for m in my_model._meta.get_fields():
        dict_fields_models[m.name] = m
    data = {}
    for key, value in my_serializer(instance, context={"request": _request}).data.items():
        if key in dict_fields_models.keys() and key != "frontend_settings" and key != "id":
            if key == "auth_token":
                data[key], _ = Token.objects.get_or_create(user=instance)
            elif dict_fields_models[key].get_internal_type() in ["FileField", "ImageField"]:
                """
                # Create a sample test file on the fly
                fpath = "testfile.txt"
                f = open(fpath, "w")
                f.write("Hello World")
                f.close()
                data[key] = open(fpath, "r")
                data[key] = document.file.open(mode='rb')
                """
                pass  # data[key] = open(value.replace("http://testserver/",""), 'rb')
            else:
                data[key] = value
            # Related objects with cascading on_delete will be deleted. we create a new related object and override the id of the deleted object
            if delete and (_field := dict_fields_models[key]):
                if (
                    _field.is_relation
                    and not _field.null
                    and _field.many_to_one
                    and (_related_fields := _field.related_fields)
                ):
                    try:
                        lh_field, rh_field = _related_fields[0]
                        if isinstance(lh_field, models.fields.AutoField) or isinstance(
                            rh_field, models.fields.AutoField
                        ):
                            data[key] = get_model_factory(dict_fields_models[key].related_model).id
                    except Exception:
                        pass
    if update or delete:
        _kwargs = {"user": superuser, "obj_factory": obj_factory}
        if delete:
            _kwargs.update({"obj_pre_deleted": instance})
        if custom_data := custom_update_data_from_factory.send(viewset, **_kwargs):
            data.update(custom_data[0][1])
    # delete object for create with post method
    if delete:
        my_model.objects.filter(pk=instance.pk).delete()
    data = {k: v for k, v in data.items() if v is not None}  # Filter out keys with `None` values
    return data


def get_kwargs(instance, viewset, request, data=None):
    """
    Get the kwargs needed for a ViewSet, handling related fields
    """
    viewset.kwargs = {"pk": instance.pk}  # set kwargs attribute to viewset
    kwargs = {"user": request.user, "profile": getattr(request.user, "profile", None), "obj_factory": instance}
    if isinstance(data, str):
        # Convert JSON string to dictionary if necessary
        data = json.loads(data)
    data = data or model_to_dict(instance)
    # Add related field IDs to kwargs
    for field in viewset.get_model()._meta.get_fields():
        if field.is_relation and field.many_to_one and (related_field_id := data.get(field.name)):
            kwargs[f"{field.name}_id"] = related_field_id

    # Add custom kwargs, if any, from signal handler
    if custom_kwargs := custom_update_kwargs.send(viewset, **kwargs):
        kwargs.update(custom_kwargs[0][1])
    return kwargs


def generate_dict_factory(factory: Factory):
    def convert_dict_from_stub(stub: StubObject) -> Dict[str, Any]:
        stub_dict = stub.__dict__
        for key, value in stub_dict.items():
            if isinstance(value, StubObject):
                stub_dict[key] = convert_dict_from_stub(value)
        return stub_dict

    def dict_factory(factory, **kwargs):
        stub = factory.stub(**kwargs)
        stub_dict = convert_dict_from_stub(stub)
        return stub_dict

    return partial(dict_factory, factory)


# check if an element of a list is present in the other
def contains(lst1, lst2):
    return any(elt in lst1 for elt in lst2)


def is_intermediate_table_m2m(model):
    if "_" in model.__qualname__ and not is_empty(model._meta.unique_together):
        list_models = [elt.lower() for elt in model.__qualname__.split("_")]
        list_models_id = [elt.lower() + "_id" for elt in model.__qualname__.split("_")]
        if (
            contains(list_models, model._meta.unique_together[0])
            and contains(list_models, model.__dict__.keys())
            and contains(list_models_id, model.__dict__.keys())
        ):
            return True
    return False


def is_empty(any_structure):
    if any_structure:
        return False
    else:
        return True


def get_factory_custom_user():
    list_models = [
        m
        for m in get_all_subclasses(models.Model)
        if m.__name__ == ("User")
        and not m.__module__.startswith(("wbcore", "django", "rest_framework", "dynamic_preferences", "eventtools"))
        and not m._meta.abstract
        and not is_intermediate_table_m2m(m)
    ]
    userfactory = [
        get_model_factory(modl) for modl in list(dict.fromkeys(list_models)) if get_model_factory(modl) is not None
    ]
    if not is_empty(userfactory):
        return userfactory[0]
    return None


def format_number(number, is_pourcent=False, decimal=2):
    number = number if number else 0
    return f'{number:,.{decimal}{"%" if is_pourcent else "f"}}'


# https://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable?page=1&tab=votes#tab-top
def datetime_converter(o):
    if isinstance(o, (datetime.date, datetime.datetime, datetime.time)):
        return o.isoformat()


def get_or_create_superuser():
    return UserFactory.create(is_active=True, is_superuser=True)



from collections import namedtuple

import pytest
from wbcore.serializers import Serializer
from wbcore.serializers.fields.number import IntegerField


class TestWBCoreSerializerFieldMixin:
    @pytest.fixture()
    def parent(self, context=None):
        if not context:
            context = {"view": object()}
        return Serializer(context=context)

    def test_read_only_callable(self):
        # check default readonly handling
        field = IntegerField(read_only=True)
        assert field.read_only is True

        field = IntegerField(read_only=False)
        assert field.read_only is False

        field = IntegerField(read_only=lambda x: True)
        assert field.read_only is True
        assert hasattr(field, "_callable_read_only")

    def test_bind_read_only_callable(self, parent):
        field = IntegerField(read_only=lambda view: False)
        field.bind("field", parent)
        assert field.read_only is False

    def test_view(self):
        view = object()
        Request = namedtuple("Request", ["parser_context"])
        request = Request({"view": view})

        c1 = {"view": view}
        c2 = {"request": request}

        field = IntegerField()
        p1 = Serializer(context=c1)
        field.bind("field", p1)
        assert field.view == view

        field = IntegerField()
        p1 = Serializer(context=c2)
        field.bind("field", p1)
        assert field.view == view

        field = IntegerField()
        p1 = Serializer()
        field.bind("field", p1)
        assert field.view is None



from django.contrib import admin
from wbcore.contrib.workflow.models import Display


@admin.register(Display)
class DisplayAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = ("id", "name", "grid_template_areas")



from dynamic_preferences.types import ModelChoicePreference


class CallableDefaultModelChoicePreference(ModelChoicePreference):
    """
    Propose a ModelChoicePreference class where default is allowed to be a callable property. This avoids unnecessary db calls at start

    This type expects a mandatory model attribute and a section of type dynamic_preferences.Section.
    """

    def __init__(self, registry=None):
        self.registry = registry
        self.queryset = self.model.objects.all()
        self.serializer = self.serializer_class(self.model)
        self._setup_signals()

    @property
    def default(self):
        raise NotImplementedError("Default property needs to be defined for this preference type")



from django.db.models import Count, IntegerField, OuterRef, Q, QuerySet, Subquery
from django.db.models.functions import Coalesce
from wbcore import viewsets
from wbcore.contrib.example_app.filters import (
    TeamResultsFilter,
    TeamResultsLeagueFilter,
)
from wbcore.contrib.example_app.models import Match, TeamResults
from wbcore.contrib.example_app.serializers import (
    TeamResultsModelSerializer,
    TeamResultsRepresentationSerializer,
)
from wbcore.contrib.example_app.viewsets.displays import (
    TeamResultsDisplayConfig,
    TeamResultsLeagueDisplayConfig,
)
from wbcore.contrib.example_app.viewsets.endpoints import TeamResultsEndpointConfig
from wbcore.contrib.example_app.viewsets.titles import (
    TeamResultsLeagueTitleConfig,
    TeamResultsTitleConfig,
)


class TeamResultsModelViewSet(viewsets.ModelViewSet):
    display_config_class = TeamResultsDisplayConfig
    queryset = TeamResults.objects.all()
    search_fields = ("team__name", "league__computed_str")
    ordering_fields = (
        "league__computed_str",
        "points",
        "match_point_difference",
        "match_points_for",
        "team__name",
        "match_points_against",
        "wins",
        "draws",
        "losses",
    )
    ordering = (
        "league__computed_str",
        "-points",
        "-match_point_difference",
        "-match_points_for",
        "team__name",
    )
    serializer_class = TeamResultsModelSerializer
    title_config_class = TeamResultsTitleConfig
    filterset_class = TeamResultsFilter
    endpoint_config_class = TeamResultsEndpointConfig

    def get_queryset(self) -> QuerySet[TeamResults]:
        games_subquery = (
            Match.objects.filter(
                Q(league=OuterRef("league"))
                & Q(Q(home=OuterRef("team")) | Q(away=OuterRef("team")))
                & Q(status=Match.MatchStatus.FINISHED)
            )
            .values("league__pk")
            .annotate(games=Count("id"))
            .values("games")
        )

        qs = (
            super()
            .get_queryset()
            .select_related("league", "team")
            .annotate(
                games_played=Coalesce(Subquery(games_subquery, output_field=IntegerField()), 0),
            )
        )
        return qs


class TeamResultsRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = TeamResults.objects.all()
    serializer_class = TeamResultsRepresentationSerializer
    search_fields = ("team__name", "league__computed_str")


class TeamResultsLeagueModelViewSet(TeamResultsModelViewSet):
    title_config_class = TeamResultsLeagueTitleConfig
    display_config_class = TeamResultsLeagueDisplayConfig
    filterset_class = TeamResultsLeagueFilter
    search_fields = ("team__name",)
    ordering_fields = (
        "points",
        "match_point_difference",
        "match_points_for",
        "team__name",
        "match_points_against",
        "wins",
        "draws",
        "losses",
    )
    ordering = (
        "-points",
        "-match_point_difference",
        "-match_points_for",
        "team__name",
    )

    def get_queryset(self) -> QuerySet[TeamResults]:
        queryset = super().get_queryset()
        if league_id := self.kwargs.get("league_id"):
            queryset = queryset.filter(league=league_id)
        return queryset



from . import operators as op
from .display import Display
from .enums import NavigationType
from .layouts import Inline, Layout, Section
from .pages import Page
from .shortcuts import create_simple_display, create_simple_section
from .operators import default
from .styles import Style
from .utils import (
    grid_definition,
    repeat,
    repeat_field,
    split_list_into_grid_template_area_sublists,
)
from .signals import add_display_pages



from celery import shared_task
from django.conf import settings
from django.core.mail.message import EmailMultiAlternatives
from django.template.loader import get_template
from django.utils.html import strip_tags
from django.utils.module_loading import import_string
from wbcore.contrib.notifications.models.notification_types import (
    NotificationTypeSetting,
)
from wbcore.contrib.notifications.models.notifications import Notification


def send_notification_email(notification: Notification):
    """Sends out a notification to the user specified inside the notification

    Args:
        notification: The notification that is going to be send
    """

    context = {
        "title": notification.title,
        "message": notification.body or "",
        "notification_share_url": notification.get_full_endpoint(as_shareable_internal_link=True),
        "notification_endpoint": notification.get_full_endpoint(),
    }
    rendered_template = get_template("notifications/notification_template.html").render(context)
    msg = EmailMultiAlternatives(
        subject=notification.title,
        body=strip_tags(rendered_template),
        from_email=getattr(settings, "WBCORE_NOTIFICATION_EMAIL_FROM", "no_reply@stainly.com"),
        to=[notification.user.email],  # type: ignore
    )
    msg.attach_alternative(rendered_template, "text/html")
    msg.send()


@shared_task
def send_notification_task(notification_pk: int):
    """A celery task to send out a notification via email, web or mobile

    Args:
        notification_pk: The primary key of the notification that is going to be send out
    """

    notification = Notification.objects.get(pk=notification_pk)

    notification_user_settings = NotificationTypeSetting.objects.get(
        notification_type=notification.notification_type,
        user=notification.user,
    )
    if notification_user_settings.enable_email:
        send_notification_email(notification)

    backend = import_string(settings.NOTIFICATION_BACKEND)
    backend.send_notification(notification)



from dataclasses import dataclass

from wbcore.enums import Operator


@dataclass(unsafe_hash=True)
class Condition:
    operator: Operator
    value: str | float | int | bool

    def __post_init__(self) -> None:
        if self.operator == Operator.EXISTS:
            assert isinstance(self.value, bool), f"{Operator.EXISTS.value} is only compatible with bool"


@dataclass(unsafe_hash=True)
class FormattingRule:
    icon: str | None = None
    style: dict | None = None
    condition: Condition | tuple | list[tuple] | None = None

    def __post_init__(self) -> None:
        assert self.icon or self.style, "icon and style cannot both be None."

    def __iter__(self):
        yield "icon", self.icon
        yield "style", self.style
        if self.condition:
            if isinstance(self.condition, tuple):
                yield "condition", self.condition
            elif isinstance(self.condition, list):
                yield "condition", [tuple(cond) for cond in self.condition]
            else:
                yield "condition", (self.condition.operator.value, self.condition.value)


@dataclass(unsafe_hash=True)
class Formatting:
    formatting_rules: list[FormattingRule]
    column: str | None = None

    def __post_init__(self) -> None:
        if self.column is None:
            assert all(
                [not bool(rule.condition) for rule in self.formatting_rules]
            ), "Specifying conditions, without a reference column is not possible."

    def __iter__(self):
        yield "column", self.column
        yield "formatting_rules", [dict(rule) for rule in self.formatting_rules]



from django.utils.translation import gettext as _
from wbcore.contrib.example_app.models import Sport
from wbcore.metadata.configs.titles import TitleViewConfig


class LeagueTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Leagues")

    def get_create_title(self) -> str:
        return _("Create League")

    def get_instance_title(self) -> str:
        return _("League")


class LeagueSportTitleConfig(LeagueTitleConfig):
    def get_create_title(self) -> str:
        if sport_id := self.view.kwargs.get("sport_id"):
            try:
                return _("Create {} League").format(Sport.objects.get(id=sport_id).name)
            except Sport.DoesNotExist:
                pass
        return super().get_create_title()



import pytest
from rest_framework.reverse import reverse
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.notifications.factories.notification_types import (
    NotificationTypeModelFactory,
)


@pytest.mark.django_db
class TestNotificationTypeRepresentationViewSet:
    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(None, 403), (["notifications.select_notificationtype"], 200)]
    )
    def test_instance(self, notification_type, client, user, status_code):
        client.force_authenticate(user)
        response = client.get(
            reverse("wbcore:notifications:notification_type_representation-detail", args=[notification_type.pk])
        )
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(None, 403), (["notifications.select_notificationtype"], 200)]
    )
    def test_list(self, client, user, status_code):
        client.force_authenticate(user)
        response = client.get(reverse("wbcore:notifications:notification_type_representation-list"))
        assert response.status_code == status_code  # type: ignore


@pytest.mark.django_db
class TestNotificationTypeSettingModelViewSet:
    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(None, 403), (["notifications.view_notificationtypesetting"], 200)]
    )
    def test_instance(self, notification_type_setting, client, user, status_code):
        client.force_authenticate(user)
        response = client.get(
            reverse("wbcore:notifications:notification_type_setting-detail", args=[notification_type_setting.pk])
        )
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(None, 403), (["notifications.view_notificationtypesetting"], 200)]
    )
    def test_list(self, client, user, status_code):
        client.force_authenticate(user)
        response = client.get(reverse("wbcore:notifications:notification_type_setting-list"))
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(None, 403), (["notifications.add_notificationtypesetting"], 403)]
    )
    def test_create(self, notification_type, client, user, status_code):
        """Regardless of permission, no user should ever be able to create notification type setting"""
        client.force_authenticate(user)
        response = client.post(
            reverse("wbcore:notifications:notification_type_setting-list"),
            {"user": user.id, "notification_type": notification_type.id},
        )
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(None, 403), (["notifications.delete_notificationtypesetting"], 403)]
    )
    def test_delete(self, notification_type_setting, client, user, status_code):
        """Regardless of permission, no user should ever be able to delete notification type setting"""
        client.force_authenticate(user)
        response = client.delete(
            reverse("wbcore:notifications:notification_type_setting-detail", args=[notification_type_setting.id])
        )
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(None, 403), (["notifications.change_notificationtypesetting"], 200)]
    )
    def test_change(self, notification_type_setting, client, user, status_code):
        notification_type_setting.user = user
        notification_type_setting.save()

        client.force_authenticate(user)
        response = client.patch(
            reverse("wbcore:notifications:notification_type_setting-detail", args=[notification_type_setting.id]),
            {"enable_web": True},
        )
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(["notifications.change_notificationtypesetting"], 200)]
    )
    def test_change_notification_type_and_user(self, notification_type_setting, client, user, status_code):
        """Regardless of the permission, you should never be able to change the user or notification type"""
        user2 = UserFactory()
        notification_type2 = NotificationTypeModelFactory()

        assert user != user2
        assert notification_type_setting.notification_type != notification_type2

        client.force_authenticate(user)
        response = client.patch(
            reverse("wbcore:notifications:notification_type_setting-detail", args=[notification_type_setting.id]),
            {"notification_type": notification_type2.pk, "user": user2.pk},  # type: ignore
        )
        assert response.status_code == status_code  # type: ignore
        assert response.data["instance"]["user"] == user.pk
        assert response.data["instance"]["notification_type"] == notification_type_setting.notification_type.pk

    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(["notifications.change_notificationtypesetting"], 404)]
    )
    def test_change_setting_from_other_user(self, notification_type_setting, client, user, status_code):
        """Regardless of permission, no user should ever be able to change a setting from a different user"""
        user2 = UserFactory()
        notification_type_setting.user = user2
        notification_type_setting.save()

        client.force_authenticate(user)
        response = client.patch(
            reverse("wbcore:notifications:notification_type_setting-detail", args=[notification_type_setting.id]),
            {"enable_web": True},
        )
        assert response.status_code == status_code  # type: ignore



from datetime import datetime
from unittest.mock import patch

import pytest
import pytz
from django.db import IntegrityError
from django.db.models import Q
from faker import Faker
from wbcore.contrib.authentication.factories import SuperUserFactory, UserFactory
from wbcore.contrib.directory.models import Person
from wbcore.contrib.example_app.factories import (
    LeagueFactory,
    MatchFactory,
    TeamFactory,
)
from wbcore.contrib.example_app.models import (
    Match,
    SportPerson,
    TeamResults,
    start_match,
)


@pytest.mark.django_db
class TestMatch:
    def test_has_permissions_superuser(self):
        user = SuperUserFactory()
        match = MatchFactory()
        assert match.has_permissions(user) is True

    def test_has_permissions_custom_perm(self):
        match = MatchFactory()
        user = UserFactory(user_permissions=["wbcore.change_match_status"])
        assert match.has_permissions(user) is True

    def test_str(self):
        match = MatchFactory(home__name="Home Team", away__name="Away Team")
        assert match.__str__() == "Home Team vs. Away Team"

    @patch("wbcore.contrib.example_app.models.start_match.apply_async")
    @patch("wbcore.contrib.example_app.models.current_app.control.revoke")
    def test_reschedule_existing_task(self, mock_revoke, mock_start):
        user = UserFactory()
        # we just need an object with a .id method
        mock_start.return_value = user
        match = MatchFactory()
        old_task_id = match.task_id
        match.reschedule_task()
        assert mock_revoke.call_args.args == (old_task_id,)
        assert mock_revoke.call_args.kwargs == {"terminate": True}
        assert mock_start.call_args.kwargs == {"eta": match.date_time, "args": [match.pk]}
        assert match.task_id == user.pk

    @patch("wbcore.contrib.example_app.models.start_match.apply_async")
    @patch("wbcore.contrib.example_app.models.current_app.control.revoke")
    def test_reschedule_new_task(self, mock_revoke, mock_start):
        user = UserFactory()
        # we just need an object with a .id method
        mock_start.return_value = user
        match = MatchFactory()
        match.task_id = None
        match.reschedule_task()
        assert mock_revoke.call_count == 1
        assert mock_start.call_args.kwargs == {"eta": match.date_time, "args": [match.pk]}
        assert match.task_id == user.pk

    def test_save_calendar_item(self):
        match = MatchFactory(league__sport__match_duration=120, date_time=datetime(2023, 5, 20, 19, 15))

        home_sport_persons = SportPerson.objects.filter(
            Q(id__in=match.home.current_players.all()) | Q(coached_team=match.home)
        )
        away_sport_persons = SportPerson.objects.filter(
            Q(id__in=match.away.current_players.all()) | Q(coached_team=match.away)
        )
        referee = SportPerson.objects.filter(id=match.referee.id)
        match_persons = home_sport_persons.union(away_sport_persons, referee)

        persons = Person.objects.filter(id__in=match_persons.values_list("profile", flat=True))
        entities_id_list = list(persons.values_list("pk", flat=True))
        entities_id_list.extend([match.home.id, match.away.id])

        assert match.period.upper == datetime(2023, 5, 20, 21, 15)
        assert set(match.entities.values_list("id", flat=True)) == set(entities_id_list)

    @patch("wbcore.contrib.example_app.models.Match.reschedule_task")
    def test_save_trigger_rescheduling(self, mock_reschedule):
        match = MatchFactory()
        match.date_time = Faker().date_time(tzinfo=pytz.utc)
        match.save()
        assert mock_reschedule.call_count == 2

    @patch("wbcore.contrib.example_app.models.Match.reschedule_task")
    def test_save_no_rescheduling(self, mock_reschedule):
        match = MatchFactory()
        match.title = "Test"
        match.save()
        assert mock_reschedule.call_count == 1

    def test_save_update_teamresults(self):
        league = LeagueFactory(points_per_win=5, points_per_draw=2, points_per_loss=1)
        home_team = TeamFactory()
        away_team = TeamFactory()

        # Home win
        match1 = MatchFactory(
            home=home_team,
            away=away_team,
            status=Match.MatchStatus.FINISHED,
            score_home=3,
            score_away=1,
            league=league,
        )
        assert set(TeamResults.objects.values_list("team__name", flat=True)) == {home_team.name, away_team.name}
        home_results = TeamResults.objects.get(team=home_team)
        away_results = TeamResults.objects.get(team=away_team)
        assert home_results.wins == 1
        assert home_results.draws == 0
        assert home_results.losses == 0
        assert home_results.form == "----W"
        assert home_results.points == 5
        assert home_results.match_points_for == 3
        assert home_results.match_points_against == 1
        assert away_results.wins == 0
        assert away_results.draws == 0
        assert away_results.losses == 1
        assert away_results.form == "----L"
        assert away_results.points == 1
        assert away_results.match_points_for == 1
        assert away_results.match_points_against == 3
        assert match1.task_id == ""

        # Home Loss
        match2 = MatchFactory(
            home=home_team,
            away=away_team,
            status=Match.MatchStatus.FINISHED,
            score_home=21,
            score_away=37,
            league=league,
        )
        home_results = TeamResults.objects.get(team=home_team)
        away_results = TeamResults.objects.get(team=away_team)
        assert home_results.wins == 1
        assert home_results.draws == 0
        assert home_results.losses == 1
        assert home_results.form == "---WL"
        assert home_results.points == 6
        assert home_results.match_points_for == 24
        assert home_results.match_points_against == 38
        assert away_results.wins == 1
        assert away_results.draws == 0
        assert away_results.losses == 1
        assert away_results.form == "---LW"
        assert away_results.points == 6
        assert away_results.match_points_for == 38
        assert away_results.match_points_against == 24
        assert match2.task_id == ""

        # Draw
        match3 = MatchFactory(
            home=home_team,
            away=away_team,
            status=Match.MatchStatus.FINISHED,
            score_home=2,
            score_away=2,
            league=league,
        )
        home_results = TeamResults.objects.get(team=home_team)
        away_results = TeamResults.objects.get(team=away_team)
        assert home_results.wins == 1
        assert home_results.draws == 1
        assert home_results.losses == 1
        assert home_results.form == "--WLD"
        assert home_results.points == 8
        assert home_results.match_points_for == 26
        assert home_results.match_points_against == 40
        assert away_results.wins == 1
        assert away_results.draws == 1
        assert away_results.losses == 1
        assert away_results.form == "--LWD"
        assert away_results.points == 8
        assert away_results.match_points_for == 40
        assert away_results.match_points_against == 26
        assert match3.task_id == ""

    @patch("wbcore.contrib.example_app.models.Match.reschedule_task")
    def test_save_no_task_id(self, mock_reschedule):
        MatchFactory(status=Match.MatchStatus.FINISHED, score_home=3, score_away=1, task_id="")
        assert not TeamResults.objects.exists()

    def test_unique_constraint(self):
        match = MatchFactory()
        with pytest.raises(IntegrityError):
            MatchFactory(home=match.home, away=match.away, date_time=match.date_time)

    def test_check_constraint(self):
        team = TeamFactory()
        with pytest.raises(IntegrityError):
            MatchFactory(home=team, away=team)

    def test_start_match_scheduled(self):
        match = MatchFactory(status=Match.MatchStatus.SCHEDULED)
        start_match(match.pk)
        assert Match.objects.get(id=match.pk).status == Match.MatchStatus.ONGOING

    def test_start_match_finished(self):
        match = MatchFactory(status=Match.MatchStatus.FINISHED)
        start_match(match.pk)
        assert Match.objects.get(id=match.pk).status != Match.MatchStatus.ONGOING



import factory
from django.contrib.auth.models import Permission
from dynamic_preferences.registries import global_preferences_registry
from wbcore.contrib.directory.factories import CompanyFactory, PersonFactory
from wbcore.contrib.directory.models import Company
from wbcore.permissions.registry import user_registry

from ..models import Group, User


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = User
        django_get_or_create = ["email"]

    email = factory.Sequence(lambda n: "email%d@admin.com" % n)
    # username = factory.Faker('user_name')
    username = factory.Sequence(lambda n: "user%d" % n)

    profile = factory.SubFactory("wbcore.contrib.directory.factories.entries.PersonFactory")
    is_active = True
    is_register = True
    is_superuser = False

    plaintext_password = factory.PostGenerationMethodCall("set_password", "defaultpassword")

    @factory.post_generation
    def user_permissions(self, create, extracted, **kwargs):
        if not create or not extracted:
            return
        for permission_key in extracted:
            app_label, codename = permission_key.split(".")
            permission = Permission.objects.get(codename=codename, content_type__app_label=app_label)
            self.user_permissions.add(permission)

    @factory.post_generation
    def groups(self, create, extracted, **kwargs):
        if not create or not extracted:
            return
        for group in extracted:
            self.groups.add(group)


class InternalUserFactory(UserFactory):
    @factory.post_generation
    def add_internal_profile(self, create, extracted, **kwargs):
        main_company_id = global_preferences_registry.manager()["directory__main_company"]
        if main_company_id and Company.objects.filter(id=main_company_id).exists():
            self.profile.employers.add(Company.objects.get(id=main_company_id))
        else:
            # Create company
            company = CompanyFactory()
            # Set global config main_company=company.id
            global_preferences_registry.manager()["directory__main_company"] = company.id
            self.profile.employers.add(company)
        self.user_permissions.add(
            Permission.objects.get(content_type__app_label="authentication", codename="is_internal_user")
        )
        user_registry.reset_cache()


class SuperUserFactory(UserFactory):
    is_superuser = True
    profile = None


class AuthenticatedPersonFactory(PersonFactory):
    user_account = factory.RelatedFactory(UserFactory, "profile")


class GroupFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("text", max_nb_chars=150)

    @factory.post_generation
    def permissions(self, create, extracted, **kwargs):
        if not create or not extracted:
            return
        for permission in extracted:
            self.permissions.add(permission)

    class Meta:
        model = Group



# TODO: Most likely to be removed
from rest_framework import serializers

from .mixins import WBCoreSerializerFieldMixin
from .number import IntegerField


class StarRatingField(IntegerField):
    field_type = "starrating"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.delimiter = ""
        self.decimal_mark = ""


class EmojiRatingField(IntegerField):
    field_type = "emojirating"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.delimiter = ""
        self.decimal_mark = ""


class RangeSelectField(WBCoreSerializerFieldMixin, serializers.FloatField):
    field_type = "rangeselect"

    def __init__(self, *args, start: float = 0.0, end: float = 1.0, step_size: float = 0.25, **kwargs):
        self.color = kwargs.pop("color", "rgb(133, 144, 162)")
        self.start = start
        self.end = end
        self.step_size = step_size
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["color"] = self.color
        representation["start"] = self.start
        representation["end"] = self.end
        representation["step_size"] = self.step_size
        return key, representation



from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class EmployerEmployeeRelationshipButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self):
        if not self.new_mode:
            if employer_id := self.view.kwargs.get("employer_id", None):
                base_url = reverse("wbcore:directory:person-list", args=[], request=self.request)
                return {
                    bt.WidgetButton(
                        endpoint=f"{base_url}?employers={employer_id}",
                        label=_("New Person"),
                        icon=WBIcon.PERSON_ADD.icon,
                        new_mode=True,
                    )
                }
            if employee_id := self.view.kwargs.get("employee_id", None):
                base_url = reverse("wbcore:directory:company-list", args=[], request=self.request)
                return {
                    bt.WidgetButton(
                        endpoint=f"{base_url}?employees={employee_id}",
                        label=_("New Company"),
                        icon=WBIcon.ADD.icon,
                        new_mode=True,
                    )
                }
        return {}



# Generated by Django 4.1.8 on 2023-04-05 09:30

import uuid

import django.db.models.deletion
import wbcore.contrib.io.models
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("django_celery_beat", "0018_improve_crontab_helptext"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="DataBackend",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("title", models.CharField(max_length=255)),
                (
                    "save_data_in_import_source",
                    models.BooleanField(
                        default=True, help_text="If true, save the data in the import_source json field"
                    ),
                ),
                (
                    "passive_only",
                    models.BooleanField(
                        default=True,
                        help_text="If True, this data backend is allowed to be called only from the import source.",
                    ),
                ),
                ("backend_class_path", models.CharField(max_length=512)),
                ("backend_class_name", models.CharField(default="DataBackend", max_length=128)),
            ],
            options={
                "verbose_name": "Data Backend",
                "verbose_name_plural": "Data Backends",
            },
        ),
        migrations.CreateModel(
            name="ImportCredential",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("key", models.CharField(max_length=255)),
                (
                    "type",
                    models.CharField(
                        choices=[
                            ("CREDENTIAL", "Credential"),
                            ("AUTHENTICATION_TOKEN", "Authentication Token"),
                            ("CERTIFICATE", "Certificate"),
                        ],
                        default="CREDENTIAL",
                        max_length=64,
                    ),
                ),
                ("username", models.CharField(blank=True, max_length=255, null=True)),
                ("password", models.CharField(blank=True, max_length=255, null=True)),
                ("authentication_token", models.CharField(blank=True, max_length=2048, null=True)),
                (
                    "certificate_pem",
                    models.FileField(
                        blank=True,
                        help_text="We are expecting a .cert file",
                        max_length=256,
                        null=True,
                        upload_to="io/import_credential/certificates",
                        validators=[wbcore.contrib.io.models.validate_pem_file],
                    ),
                ),
                (
                    "certificate_key",
                    models.FileField(
                        blank=True,
                        help_text="We are expecting a .key file",
                        max_length=256,
                        null=True,
                        upload_to="io/import_credential/certificates",
                        validators=[wbcore.contrib.io.models.validate_key_file],
                    ),
                ),
                ("additional_resources", models.JSONField(blank=True, default=dict, null=True)),
                ("validity_start", models.DateTimeField(blank=True, null=True)),
                ("validity_end", models.DateTimeField(blank=True, null=True)),
            ],
            options={
                "verbose_name": "Import Credential",
                "verbose_name_plural": "Import Credentials",
            },
        ),
        migrations.CreateModel(
            name="ParserHandler",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("parser", models.CharField(max_length=255)),
                ("handler", models.CharField(max_length=255)),
                (
                    "allow_file_type",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("text/csv", "CSV"),
                            ("application/json", "JSON"),
                            ("application/vnd.ms-excel", "XLS"),
                            ("application/vnd.openxmlformats-officedocument.spreadsheetml.sheet", "XLSX"),
                        ],
                        max_length=68,
                        null=True,
                    ),
                ),
            ],
            options={
                "verbose_name": "Parser-Handler",
                "verbose_name_plural": "Parsers-Handlers",
                "unique_together": {("parser", "handler")},
            },
        ),
        migrations.CreateModel(
            name="Provider",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("title", models.CharField(max_length=255)),
                ("key", models.CharField(max_length=255, unique=True)),
            ],
            options={
                "verbose_name": "Provider",
                "verbose_name_plural": "Providers",
            },
        ),
        migrations.CreateModel(
            name="Source",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("title", models.CharField(max_length=128)),
                ("uuid", models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ("is_active", models.BooleanField(default=True)),
                ("connection_parameters", models.JSONField(blank=True, default=dict, null=True)),
                ("import_parameters", models.JSONField(blank=True, default=dict, null=True)),
                ("import_timedelta_interval", models.IntegerField(default=0)),
                (
                    "credentials",
                    models.ManyToManyField(blank=True, related_name="sources", to="io.importcredential"),
                ),
                (
                    "crontab",
                    models.ForeignKey(
                        blank=True,
                        help_text="Crontab Schedule to run the task on.  Set only one schedule type, leave the others null.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="django_celery_beat.crontabschedule",
                        verbose_name="Crontab Schedule",
                    ),
                ),
                (
                    "data_backend",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="io.databackend",
                        verbose_name="Data Backend",
                    ),
                ),
                (
                    "parser_handler",
                    models.ManyToManyField(
                        related_name="sources", to="io.parserhandler", verbose_name="Parser/Handler"
                    ),
                ),
                (
                    "periodic_task",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="sources",
                        to="django_celery_beat.periodictask",
                        verbose_name="Periodic Task",
                    ),
                ),
            ],
            options={
                "verbose_name": "Source",
                "verbose_name_plural": "Sources",
            },
        ),
        migrations.CreateModel(
            name="ImportSource",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("last_updated", models.DateTimeField(auto_now=True)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("PENDING", "Pending"),
                            ("IMPORTED", "Imported"),
                            ("ERROR", "Error"),
                            ("IGNORED", "Ignore"),
                        ],
                        default="PENDING",
                        max_length=16,
                    ),
                ),
                (
                    "file",
                    models.FileField(blank=True, max_length=256, null=True, upload_to="io/import_source/files"),
                ),
                ("data", models.JSONField(blank=True, default=dict, null=True)),
                (
                    "save_data",
                    models.BooleanField(default=True, help_text="If True, will save the raw data in a field"),
                ),
                ("progress_index", models.PositiveIntegerField(default=0)),
                ("log", models.TextField(blank=True, null=True)),
                ("origin", models.CharField(blank=True, max_length=255, null=True)),
                (
                    "parser_handler",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="io.parserhandler"),
                ),
                (
                    "source",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="io.source",
                        verbose_name="Source",
                    ),
                ),
            ],
            options={
                "verbose_name": "Import Source",
                "verbose_name_plural": "Import Sources",
            },
        ),
        migrations.CreateModel(
            name="ImportedObjectProviderRelationship",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
                ("object_id", models.PositiveIntegerField()),
                ("provider_identifier", models.CharField(max_length=255)),
                (
                    "content_type",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="contenttypes.contenttype"),
                ),
                (
                    "provider",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="imported_object_relationships",
                        to="io.provider",
                    ),
                ),
            ],
            options={
                "verbose_name": "Content object Provider Identifier relationship",
                "verbose_name_plural": "Content object Provider Identifier relationships",
            },
        ),
        migrations.AddField(
            model_name="databackend",
            name="provider",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="io.provider",
                verbose_name="Provider",
            ),
        ),
        migrations.AddIndex(
            model_name="importedobjectproviderrelationship",
            index=models.Index(
                fields=["content_type", "object_id", "provider"], name="io_imported_content_18b6a8_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="importedobjectproviderrelationship",
            index=models.Index(
                fields=["content_type", "provider_identifier", "provider"],
                name="io_imported_content_3da41e_idx",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="importedobjectproviderrelationship",
            unique_together={
                ("content_type", "provider_identifier", "provider"),
                ("content_type", "object_id", "provider"),
            },
        ),
    ]



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.workflow.models import Process, ProcessStep
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default


class ProcessDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="workflow", label=_("Workflow")),
            dp.Field(key="id", label=_("UUID")),
            dp.Field(key="started", label=_("Started")),
            dp.Field(key="finished", label=_("Finished")),
            dp.Field(key="content_type", label=_("Attached Model")),
        ]

        if "workflow_id" in self.view.kwargs:
            fields.pop(0)

        color_map = Process.ProcessState.get_color_map()

        return dp.ListDisplay(
            fields=fields,
            legends=get_state_legend(color_map),
            formatting=get_state_formatting(color_map),
        )

    def get_instance_display(self) -> Display:
        process_steps_section = Section(
            key="process_steps_section",
            collapsible=True,
            title=_("Process Steps"),
            display=Display(
                pages=[
                    Page(
                        title=_("Process Steps"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["process_steps_inline"]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="process_steps_inline", endpoint="process_steps")],
                            )
                        },
                    ),
                ]
            ),
        )
        return Display(
            pages=[
                Page(
                    title=_("Process"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["state", "."],
                                ["workflow", "id"],
                                ["started", "finished"],
                                ["content_type", "instance_id"],
                                ["process_steps_section", "process_steps_section"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                            sections=[process_steps_section],
                        ),
                    },
                ),
            ]
        )


def get_state_legend(color_map: list[tuple]) -> list[dp.Legend]:
    """Dynamically creates the process or process step legend based on the state enum using the color mapping"""

    legend_items = []
    for state, color in color_map:
        legend_items.append(dp.LegendItem(icon=color, label=state.label, value=state.value))
    return [dp.Legend(key="state", items=legend_items)]


def get_state_formatting(color_map: list[tuple]) -> list[dp.Formatting]:
    """Dynamically creates the process or process step list formatting based on the state enum using the color mapping"""

    formatting_rules = []
    for state, color in color_map:
        formatting_rules.append(dp.FormattingRule(condition=("==", state.value), style={"backgroundColor": color}))
    return [dp.Formatting(column="state", formatting_rules=formatting_rules)]


class ProcessStepDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="id", label=_("UUID")),
            dp.Field(key="step", label=_("Step")),
            dp.Field(key="process", label=_("Process")),
            dp.Field(key="started", label=_("Started")),
            dp.Field(key="finished", label=_("Finished")),
            dp.Field(key="error_message", label=_("Error Message")),
            dp.Field(key="assignee", label=_("Assignee")),
            dp.Field(key="group", label=_("Group")),
            dp.Field(key="permission", label=_("Permission")),
            dp.Field(key="status", label=_("Status")),
        ]

        if "process_id" in self.view.kwargs:
            fields.pop(2)
        elif "step_id" in self.view.kwargs:
            fields.pop(1)

        color_map = ProcessStep.StepState.get_color_map()

        return dp.ListDisplay(
            fields=fields,
            legends=get_state_legend(color_map),
            formatting=get_state_formatting(color_map),
        )

    def get_instance_display(self) -> Display:
        grid_fields = [
            ["state", ".", "."],
            ["id", "step", "process"],
            ["started", "finished", "status"],
            ["assignee", "group", "permission"],
            ["error_message", "error_message", "."],
        ]
        # if "pk" in self.view.kwargs:
        #     process_step: ProcessStep = self.view.get_object()
        #     if (step := process_step.step).step_type == Step.StepType.USERSTEP and (
        #         display := step.get_casted_step().display
        #     ):
        #         # Display those fields declared in the user step display model
        #         grid_fields = display.grid_template_areas
        #     elif attached_data := process_step.process.workflow.attached_data.all():
        #         # Display the fields from attached data objects
        #         # TODO: Optimize
        #         sublists = split_list_into_grid_template_area_sublists(
        #             [f"data__{str(data.pk)}" for data in attached_data], 3
        #         )
        #         grid_fields = grid_fields + sublists

        return Display(
            pages=[
                Page(
                    title=_("Process Step"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=grid_fields,
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                    },
                ),
            ]
        )


class AssignedProcessStepDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=(
                dp.Field(key="step", label=_("Step")),
                dp.Field(key="attached_model", label=_("Attached Model")),
                dp.Field(key="workflow_name", label=_("Workflow")),
                dp.Field(key="started", label=_("Started")),
                dp.Field(key="finished", label=_("Finished")),
                dp.Field(key="group", label=_("Group")),
                dp.Field(key="permission", label=_("Permission")),
                dp.Field(key="status", label=_("Status")),
            )
        )



from django.http import HttpRequest
from wbcore.contrib.directory.models import Entry


class EntryPermissionMixin:
    queryset = Entry.objects.all()
    request: HttpRequest

    def get_queryset(self):
        # allow the user to see only entries they can see
        return super().get_queryset().filter_for_user(self.request.user)  # type: ignore



from itertools import product
from string import ascii_uppercase

import factory

from .models import Geography

POSSIBLE_COUNTRY_CODE_2 = ["".join(i) for i in product(ascii_uppercase, repeat=2)]
POSSIBLE_COUNTRY_CODE_3 = ["".join(i) for i in product(ascii_uppercase, repeat=3)]


class ContinentFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Geography

    name = factory.Faker("text", max_nb_chars=10)


class CountryFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Geography

    code_3 = factory.Sequence(lambda n: POSSIBLE_COUNTRY_CODE_3[n % len(POSSIBLE_COUNTRY_CODE_3)])
    code_2 = factory.Sequence(lambda n: POSSIBLE_COUNTRY_CODE_2[n % len(POSSIBLE_COUNTRY_CODE_2)])
    parent = factory.LazyAttribute(lambda x: ContinentFactory.create())
    name = factory.Faker("country")


class StateFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Geography

    code_3 = factory.Sequence(lambda n: POSSIBLE_COUNTRY_CODE_3[n % len(POSSIBLE_COUNTRY_CODE_3)])
    code_2 = factory.Sequence(lambda n: POSSIBLE_COUNTRY_CODE_2[n % len(POSSIBLE_COUNTRY_CODE_2)])
    parent = factory.SubFactory(CountryFactory)
    name = factory.Faker("state")


class CityFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Geography

    code_3 = factory.Sequence(lambda n: POSSIBLE_COUNTRY_CODE_3[n % len(POSSIBLE_COUNTRY_CODE_3)])
    code_2 = factory.Sequence(lambda n: POSSIBLE_COUNTRY_CODE_2[n % len(POSSIBLE_COUNTRY_CODE_2)])
    parent = factory.SubFactory(StateFactory)
    name = factory.Faker("city")



def lt(value: int) -> str:
    return _stringify("<", value)


def lte(value: int) -> str:
    return _stringify("<=", value)


def gt(value: int) -> str:
    return _stringify(">", value)


def gte(value: int) -> str:
    return _stringify(">=", value)


def default() -> str:
    return "default"


def _stringify(operator: str, value: int) -> str:
    return f"{operator}{value}"



from .document_types import DocumentTypeModelDisplay
from .documents import DocumentModelDisplay
from .shareable_links import ShareableLinkAccessModelDisplay, ShareableLinkModelDisplay
from .document_model_relationships import DocumentModelRelationshipViewConfig



from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, MenuItem

DMS_MENUITEM = MenuItem(
    label=_("Documents"),
    endpoint="wbcore:documents:document-list",
    permission=ItemPermission(permissions=["documents.view_document"]),
    add=MenuItem(
        label=_("Create New Document"),
        endpoint="wbcore:documents:document-list",
        permission=ItemPermission(permissions=["documents.add_document"]),
    ),
)

DOCUMENT_TYPE_MENUITEM = MenuItem(
    label=_("Document Types"),
    endpoint="wbcore:documents:documenttype-list",
    permission=ItemPermission(permissions=["documents.view_documenttype"]),
    add=MenuItem(
        label=_("Create New Document Type"),
        endpoint="wbcore:documents:documenttype-list",
        permission=ItemPermission(permissions=["documents.add_documenttype"]),
    ),
)



from wbcore import viewsets
from wbcore.contrib.example_app.filters import RoleFilter
from wbcore.contrib.example_app.models import Role
from wbcore.contrib.example_app.serializers import (
    RoleModelSerializer,
    RoleRepresentationSerializer,
)
from wbcore.contrib.example_app.viewsets.displays import RoleDisplayConfig
from wbcore.contrib.example_app.viewsets.titles import RoleTitleConfig


class RoleModelViewSet(viewsets.ModelViewSet):
    display_config_class = RoleDisplayConfig
    queryset = Role.objects.all()
    serializer_class = RoleModelSerializer
    title_config_class = RoleTitleConfig
    ordering_fields = ordering = ("title",)
    filterset_class = RoleFilter
    search_fields = ("title",)


class RoleRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Role.objects.all()
    serializer_class = RoleRepresentationSerializer
    search_fields = ("title",)



from collections import defaultdict

from .base import WBCoreViewConfig


class FieldsViewConfig(WBCoreViewConfig):
    metadata_key = "fields"
    config_class_attribute = "fields_config_class"

    def get_metadata(self) -> dict:
        fields = defaultdict(dict)
        if (serializer_class := getattr(self.view, "get_serializer", None)) and (serializer := serializer_class()):
            for field_name, field in serializer.fields.items():
                field_key, field_representation = field.get_representation(self.request, field_name)
                fields[field_key].update(field_representation)

        return fields



from .components import (
    SecondStart,
    SecondEnd,
    MinuteStart,
    MinuteEnd,
    HourStart,
    HourEnd,
    WeekStart,
    WeekEnd,
    MonthStart,
    MonthEnd,
    QuarterStart,
    QuarterEnd,
    YearStart,
    YearEnd,
    Now,
)
from .offsets import Second, Minute, Hour, Day, BusinessDay, Week, Month, Quarter, Year



import urllib.parse
from collections import defaultdict
from typing import Any


def nest_row(row: dict[str, Any], sep="__") -> dict[str, Any]:
    """
    Utility function to nest row based on seperator.

    Examples:
        Row with key containing "__" such as "key__nested_key" will be converted to a dictionary "{"key": "nested_key"}"
    Args:
        row: The un-nested row
        sep: The seperated to use for nested separation

    Returns:
        A nested dictionary
    """
    res = defaultdict(dict)
    for k, v in row.items():
        splits = k.split(sep, maxsplit=1)
        if len(splits) == 2:
            res[splits[0]][splits[1]] = v
        else:
            res[k] = v
    return dict(res)


def parse_endpoint(request, endpoint: str, **extra_kwargs) -> str:
    parse_url = urllib.parse.urlsplit(urllib.parse.unquote(request.get_full_path()))
    endpoint = parse_url.path + endpoint + "/"
    params = extra_kwargs | dict(urllib.parse.parse_qsl(parse_url.query))
    if params:
        params_repr = "&".join([f"{k}={v}" for k, v in params.items()])
        endpoint += f"?{params_repr}"
    return endpoint


def get_import_export_identifier(view) -> str:
    try:
        identifier = view.IDENTIFIER
    except AttributeError:
        identifier = "{0.app_label}:{0.model}".format(view.get_content_type())
    return identifier
